# File: evm_AAA_Common.tcl
# Purpose: EVM common procedures (utilities)
#

# $Id: evm_AAA_Common.tcl 328569 2019-12-10 13:36:49Z smarteau $

## Standard informational messages for error/waiver conditions - DO NOT EDIT !
set evmWaiver(errMsg) "\[This ERROR can be eliminated by means of a Waiver.\nSee http://www.eso.org/sci/observing/phase2/SMGuidelines/WaiverChanges.html for more details.\]\n"
set evmWaiver(wavMsg) "\[This ERROR is being ignored because of your successful Phase 2 Waiver Request.\]\n"

# Convert a full run ID to integer
# E.g. : 077.A-0123(C) --> 7701232
#
# inputs:
#  - runID: run ID in the full form "077.A-0123(C)"
#
# output:
#  - run ID in its integer form (or -1 if input does
#    not have valid format)
#
proc convertRunID { runID } {

    # Regular expression for full run ID
    set runRegexp {^([0-5][1-9]?[0-9][0-9]|60)\.[A-Z]-[0-9][0-9][0-9][0-9]\([A-Z]\)$}

    # If the input does not match the expected format,
    # exit procedure (return -1)
    if { ![regexp $runRegexp $runID] } {
	return -1
    }

    # Remove panel letter, parentheses, dots, dashes
    regsub -all {\.[A-Z]-|[()]} $runID "" id

    # Get run letter from end of string
    set letter [string range $id end end]
    # Get run ID without run letter

    set id [string range $id 0 end-1]

    set propNumber [string range $id end-3 end]

    set prefix [string range $id 0 end-4]

    if { [string length $prefix] >= 3 } {
	set prefix [string range $prefix 1 end]
    }

    # Transform letter into index (A=00, B=01, C=02, etc.)
    scan $letter "%c" ascii
    set letter_id [format "%02d" [expr $ascii - 65 ] ]

    # Return run id in integer format
    return "$prefix$propNumber$letter_id"
}

# Parses a single line of a PAF file
#
# input: a string representing a line of a PAF file
# output: a list of two elements containning the PAF keyword
#         and the associated value
#
proc parsePAFLine { string } {

    # Regular expression that matches a line of a PAF file
    set pafPattern "^((\[A-Z0-9\-\]+\\.)+\[A-Z0-9\-\]+)(\[\\ \\t\]+\"(.*)\")?;?(\[\\ \\t\]*\#?.*)?$"

    # Simpler regular expression that sometimes also matches a line of a PAF file !
    set pafSimplerPattern "^((\[A-Z0-9\-\]+\\.)+\[A-Z0-9\-\]+)(\[\\ \\t\]+(\[^;\]*))?;?$"

    set keyword ""
    set value ""

    # Parse the line and retrieve the keyword and value
    if { [regexp $pafPattern $string dummy1 keyword dummy2 dummy3 value] != 0 } {

	# If successful, return keyword and value
	return [list $keyword $value]

    } elseif { [regexp $pafSimplerPattern $string dummy1 keyword dummy2 dummy3 value] != 0 } {

	return [list $keyword $value]

    } else {

	# ... else return empty list.
	return {}

    }

}

# Procedure: formatPrintedTime
#
# Formats time in h m s format
# arguments:
#  - etimesecs: time in seconds
#
# returns:
#  - printedTime: time string formatted as h m s
#
proc formatPrintedTime { etimesecs } {
    set etime [expr  int($etimesecs) ]
    set hours [expr $etime / 3600 ]
    set minutes [expr ($etime % 3600) / 60]
    set seconds [expr ($etime % 3600) % 60]

    set printedTime [format "%dh%02dm%02ds" $hours $minutes $seconds]
    return $printedTime
} ; # End proc formatPrintedTime

#############################################################
## Auxiliary procedures for arrays
#############################################################

# getArrayElements
#
# Inputs:
#  - arrayName : name of the array to search
#  - pattern   : global search pattern on array indices (default = *)
#
# Output:
#  List of elements of the array that match the pattern
#
proc getArrayElements { arrayName { pattern "*" } } {

    # work on variable at upper level
    upvar $arrayName localArray

    # init list of matching elements
    set arrayElements {}

    # Get all indexes from the array (with the possible
    # restriction of a given pattern)
    set indexes [array names localArray $pattern]

    # If no matching indexes are found
    if { [llength $indexes] == 0 } {

	# return the empty list
	return $arrayElements
    }

    # Loop on all array indexes
    foreach idx $indexes {

	# Append corresponding element to list of results
	lappend arrayElements $localArray($idx)

    }; # end loop

    # Return list of elements
    return $arrayElements
}

# getArrayIndex
#
# Inputs:
#  - arrayName : name of the array to search
#  - pattern   : global search pattern on array elements
#
# Output:
#  List of indexes of the array that match the element pattern
#
proc getArrayIndex { arrayName pattern } {

    # Work on variable at upper level
    upvar $arrayName localArray

    # Init list of matching indexes
    set indexList {}

    # Get sorted list of indexes of the array
    set indexes [lsort [array names localArray]]

    # Loop on indexes
    foreach idx $indexes {

	# If the value of the element of the array with the
	# current index matches the patternm,
	if { [string match $pattern $localArray($idx)] == 1 } {

	    # append the current index to the list of results
	    lappend indexList $idx
	}
    }

    # Return matching indexes
    return $indexList
}

##################################################
# Common mathematical utilities
##################################################

# max:
#
# simple procedure that returns the maximum of two
# values.
#
proc max { arg1 arg2 } {

    if { $arg1 > $arg2 } {

	return $arg1

    } else {

	return $arg2

    }
}

# min:
#
# simple procedure that returns the minimum of two
# values.
#
proc min { arg1 arg2 } {

    if { $arg1 < $arg2 } {

	return $arg1

    } else {

	return $arg2

    }
}

##################################################
# Tcl list utilities
##################################################

# uniq:
#
# suppresses duplicate elements from a list
#
# Inputs: list (list of elements to clean from
#               duplicates)
#
proc uniq { list } {

    # Get number of elements in the list
    set n [llength $list]

    # If list is empty,
    if { $n == 0 } {

	# Return empty list
	return {}

    }

    # Init clean list
    set uList {}

    # Loop on all elements
    foreach i $list {

	# If current element is not found in the
	# clean list
	if { [lsearch -exact $uList $i] == -1 } {

	    # Add the current element to the clean list
	    lappend uList $i

	}
    }

    # Return clean list
    return $uList

}

# intersect:
#
# returns a list of common values between two lists
#
# Inputs: list1 and list2
#
proc intersect { list1 list2 } {

    set intersection [list]

    foreach elem $list1 {
	if {[lsearch $list2 $elem] != -1} {
	    lappend intersection $elem
	}
    }

    return $intersection
}

###############################################
# Coordinates utilities
###############################################

# ra2arcsec
#
# Procedure: convert RA string to arcseconds. The RA is passed
#            as a decimal number representing hhmmss.ddd
#
# inputs:
#   - coord : the RA in the form of a decimal number, hhmmss.ddd
#
# output:
#  returns the RA in arcseconds
#
proc ra2arcsec { coord } {

    # coordinate is put in the form +/-h[h]mmss.ddd
    set newcoord [format "%+011.3f" $coord]

    # scan the coordinate string to extract hours, minutes and seconds
    regexp {([+-][0-9][0-9]?)([0-9][0-9])([0-9][0-9].[0-9][0-9][0-9])} \
	$newcoord dummy hours minutes seconds

    # Get sign of RA
    if { [string index $newcoord 0] == "+" } {
	set sign 1
    } else {
	set sign -1
    }

    # Erase extraneous signs and leading zeros to avoid
    # Tcl misinterpretation (numbers with leading zeros are
    # octal in Tcl)
    regsub {^[+-][0]?} $hours "" hours
    regsub {^[0]?} $minutes "" minutes
    regsub {^[0]?} $seconds "" seconds

    # Compute number of arcseconds (hours ==> *15)
    set arcsec_val [expr $sign * 15 * (abs($hours) * 3600 + \
	$minutes * 60. + $seconds) ]

    return $arcsec_val
}

# dec2arcsec
#
# Procedure: convert DEC string to arcseconds. The DEC is passed
#            as a decimal number representing ddmmss.ddd
#
# inputs:
#   - coord : the DEC in the form of a decimal number, ddmmss.ddd
#
# output:
#  returns the DEC in arcseconds
#
proc dec2arcsec { coord } {

    # coordinate is put in the form +/-ddmmss.d
    set newcoord [format "%+011.3f" $coord]

    # scan the coordinate string to extract degrees, minutes and seconds
    regexp {([+-][0-9][0-9]?)([0-9][0-9])([0-9][0-9].[0-9][0-9][0-9])} \
	$newcoord dummy degrees minutes seconds

    # Get sign of DEC
    if { [string index $newcoord 0] == "+" } {
	set sign 1
    } else {
	set sign -1
    }

    # Erase extraneous signs and leading zeros to avoid
    # Tcl misinterpretation (numbers with leading zeros are
    # octal in Tcl)
    regsub {^[+-][0]?} $degrees "" degrees
    regsub {^[0]?} $minutes "" minutes
    regsub {^[0]?} $seconds "" seconds

    # Compute number of arcseconds
    set arcsec_val [expr $sign * (abs($degrees) * 3600 + \
	$minutes * 60. + $seconds) ]

    return $arcsec_val
}

# computeDistance
#
# Procedure: computes the angular distance between two points
#            of the sky
#
# inputs:
#   - ra1 : right ascension of the first point, given in arcseconds
#   - ra2 : right ascension of the second point, given in arcseconds
#   - dec1: declination of the first point, given in arcseconds
#   - dec2: declination of the second point, given in arcseconds
#
# output:
#  returns the angular distance in arcseconds
#
proc computeDistance { ra1 ra2 dec1 dec2 } {

    # Conversion factor from arcseconds to radians
    set arcsec2rad [expr 3.14159265358979 / (3600. *180.)]

    # convert input coordinates to radians
    set ra1 [expr $ra1 * $arcsec2rad]
    set ra2 [expr $ra2 * $arcsec2rad]
    set dec1 [expr $dec1 * $arcsec2rad]
    set dec2 [expr $dec2 * $arcsec2rad]

    # Compute angular distance:
    # cos(theta) = sin(dec1)*sin(dec2) + cos(dec1)*cos(dec2)*cos(ra2-ra1)
    #

    set t [expr sin($dec1)*sin($dec2) + cos($dec1)*cos($dec2)*cos($ra2-$ra1)]
    set dist [expr acos($t)]

    # Convert to arcseconds
    set dist [expr abs($dist / $arcsec2rad)]

    return $dist
}

# Procedure: angularSeparationCheck
#
# Checks for angular separation between targets for OBs in a Concatenation
# arguments:
#  - ra_dec_list: list of ra,dec pairs for targets
#
# returns:
#  - msg: error message or blank string
#
proc angularSeparationCheck { ra_dec_list } {

    #get number of ra,dec pairs
    set np [llength $ra_dec_list]
    set msg ""
    set maxSep 0
    set lim75 [format "%.2f" [expr 75 * 3600]]
    set lim45 [format "%.2f" [expr 45 * 3600]]

    for {set i 0} {$i < [expr $np - 1]} {incr i} {
	set ra_dec1 [lindex $ra_dec_list $i]
	set ra1 [lindex $ra_dec1 0]
	set dec1 [lindex $ra_dec1 1]

	for {set j [expr $i + 1]} {$j < $np} {incr j} {
	    set ra_dec2 [lindex $ra_dec_list $j ]
	    set ra2 [lindex $ra_dec2 0]
	    set dec2 [lindex $ra_dec2 1]

	    set maxSep [format "%.2f" [computeDistance $ra1 $ra2 $dec1 $dec2]]

	    #return the moment it hits lower limit of 45
	    if { $maxSep > $lim45 } {
		return $maxSep
	    }
	}; #End j loop
    }; #End i loop
    return $maxSep

} ; # End proc angularSeparationCheck

# FindEnclosingEntries --
#    Search within a sorted list
#
# Arguments:
#    val      Value to be searched
#    values   List of values to be examined
#
# Results:
#    Returns a list of the previous and next indices
#
proc FindEnclosingEntries { val values } {
    set found 0
    set row2  0
    foreach v $values {
	if { $val <= $v } {
	    set row   [expr {$row2-1}]
	    set found 1
	    break
	}
	incr row2
    }

    #
    # Border cases: extrapolation needed
    #
    if { ! $found } {
	incr row2 -1
	set  row $row2
    }
    if { $row == -1 } {
	set row $row2
    }

    return [list $row $row2]
}

# lin_interp
#
# simple linear 1-D interpolation
#
# args: table = name of table containing values
#               values are extracted from the index
#               of the table, considering only the characters
#               before the first comma.
#               Example: index="aaa,bb,1" --> value=aaa
#           x = input value
#
# returns: value based on table, interpolated at x
#
proc lin_interp { table x_val } {

    upvar $table table_l
    set vals [array names table_l]

    set valRange {}

    foreach v $vals {
	regexp {^[^,]+} $v x
	lappend valRange $x
    }

    set valRange [lsort -real $valRange]

    if { [lindex $valRange 0] > $x_val } {
	return $table_l([array names table_l "[lindex $valRange 0],*"])
    }

    if { [lindex $valRange end] < $x_val } {
	return $table_l([array names table_l "[lindex $valRange end],*"])
    }

    foreach {row_min row_max} [FindEnclosingEntries $x_val $valRange] break

    set x_min [lindex $valRange $row_min]
    set x_max [lindex $valRange $row_max]

    set lowIdx [array names table_l "$x_min,*"]
    set highIdx [array names table_l "$x_max,*"]

    set y_min $table_l($lowIdx)
    set y_max $table_l($highIdx)

    if { $x_min != $x_val } {
	set y_val [expr {$y_min+($y_max-$y_min)*($x_val-$x_min)/($x_max-$x_min)}]
    } else {
	set y_val $y_min
    }

    return $y_val
}

###############################################
# LST and time intervals utilities
###############################################

# Procedure: unionInt
#
# Sorts list of startTime endTime LST pairs for OBs in a concatenation, the order of the arguments
# is very important. The lower bound returned is always the value of x1 of the top OB, and the upper
# bound depends on the longer interval.
# arguments:
#  - x1 and y1: are time interval start and end times for the previous OB in a concatenation
#  - x2 and y2: are time interval start and end times for the later OB in a concatenation
#
# returns:
#  - 0, 1, -1 depending of equal, greater than or less than
#
proc unionInt { x1 y1 x2 y2 } {
    #puts "x1 y1 x2 y2 $x1 $y1 $x2 $y2"
    if { $y2 < $x1 || $x2 > $y1} {
	return [list 0 0]
    }
    return [list $x1 $y2]
}

# Procedure: obTIsort
#
# Sorts list of startTime endTime pairs
# arguments:
#  - etimesecs: time in seconds
#
# returns:
#  - 0, 1, -1 depending of equal, greater than or less than
#
proc obTIsort {t1 t2} {
    set startT1 [lindex $t1 0]
    set startT2 [lindex $t2 0]
    if { [expr $startT1 - $startT2] == 0 } {
	return 0
    } else {
	return [expr int(($startT1 - $startT2)/abs($startT1 - $startT2))]
    }

} ; # End proc obTIsort

# Procedure: wrapList
# arguments:
#  - list: list of elements
#  - length: desired length
#
# returns: list of desired length based on provided list. The original
# list is either truncated or "wrapped around" to produce a list of the
# desired length.
proc wrapList { list length } {

    set l [llength $list]

    if { $length <= $l } {

	return [lrange $list 0 [expr $length - 1]]

    } else {

	set nbRepeat [expr $length / $l]
	set list2 $list
	for { set i 1 } { $i <= $nbRepeat } {incr i} {
	    set list2 [concat $list2 $list]
	}

	return [lrange $list2 0 [expr $length - 1]]

    }

} ; # End proc wrapList

# Set of 3 procedures to compute whether 2 LST intervals overlap.
proc LST_intersect {b as ae} {
    if { $as > $ae } {
	if { $b >= $as || $b <= $ae } {
	    return true
	}
    } else {
	if { $b >= $as && $b <= $ae } {
	    return true
	}
    }
    return false
}

proc normalize24 { h } {
    return [expr $h - 24* int($h/24)]
}

proc LST_overlap { l1_start l1_end l2_start l2_end } {
	
    # First: normalize to 24
    set l1s_norm [normalize24 $l1_start]
    set l1e_norm [normalize24 $l1_end]
    set l2s_norm [normalize24 $l2_start]
    set l2e_norm [normalize24 $l2_end]
					
    # Check if the 2 intervals intersect
    return [expr [LST_intersect $l2s_norm $l1s_norm $l1e_norm] || [LST_intersect $l1s_norm $l2s_norm $l2e_norm] ]
	
}