# File: evmBBB_OMEGACAMtests.tcl
# Purpose: OMEGACAM specific EVM tests
# Author: Stephane Marteau
#

# $Id: evm_BBB_OMEGACAMtests.tcl 328750 2019-12-13 14:03:01Z smarteau $

# OB data - global array containing values of all needed parameters for all templates
# of all OBs.
#
# OBdata(<ob_id>,<tpl_idx>,<keyword>) = <value>
#

global OBdata
array set OBdata {}

#####################  OMEGACAM TEMPLATES  ####################
global evmOMEGACAMTpl
array set evmOMEGACAMTpl {}

# OMEGACAM acquisition templates
set evmOMEGACAMTpl(acq,1)        OMEGACAM_img_acq_movetogap
set evmOMEGACAMTpl(acq,2)        OMEGACAM_img_acq

# OMEGACAM science templates
set evmOMEGACAMTpl(sci,1)        OMEGACAM_img_obs_dither
set evmOMEGACAMTpl(sci,2)        OMEGACAM_img_obs_jitter
set evmOMEGACAMTpl(sci,3)        OMEGACAM_img_obs_offset
set evmOMEGACAMTpl(sci,4)        OMEGACAM_img_obs_stare

#############################################################
##
## Main EVM Procedures
##
#############################################################

#### OMEGACAM Level 1 Checks - General tests

proc checkOMEGACAMLevel1 { object } {

    # Global variables
    global scriptingBox
    global OBdata Phase1data

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    # Run letter
    if { ![regexp {([A-Z])\)$} $runID dummy runLetter] } {
	regexp {[0-9][0-9][0-9]$} $runID runLetter
	set runLetter [format "%c" [expr [string trimleft $runLetter 0] + 64]]
    }

    #
    # Checks at container level
    #

    # For VM: check that containers are not used [General Check]
    if { $obRunMode == "v" && $isContainer && $objSelected } {
	set msg "It is not possible to use OB containers in Visitor Mode!"
	issueError $containerID $msg $runID    
    }

    # Sanity check for time delays in TIME-LINK containers [General Check]
    if { $obRunMode == "s" && $objType == "TIME_LINK"  && $objSelected } {

	set earliestStartTime 0
	set earliestEndTime 0
	set hasAbsoluteTime false

	set sumEaP 0

	for {set i 0} {$i < [llength $listOfObsBlocks]} {incr i} {

	    set ob [$objRef getObsBlock $i]
	    set ob_id [evmID $ob]
	    set obName $OBdata($ob_id,name)

	    # for the first OB get T1 and T2 if defined
	    if { $i == 0 } {
		# time interval set
		set tis [evmTimeIntervals $ob]
		if { [llength $tis] } {
		    # absolute time intervals defined for the first OB
		    set hasAbsoluteTime true
		    set tt 0
		    # in case the returned set of time intervals in not ordered, get the earliest
		    foreach ti $tis {
			set startTime [evmStartTime $ti]
			set endTime [evmEndTime $ti]
			if { $tt == 0 } {
			    set earliestStartTime $startTime
			    set earliestEndTime $endTime
			} elseif { $startTime < $earliestStartTime } {
			    set earliestStartTime $startTime
			    set earliestEndTime $endTime
			}
			incr tt
		    }
		} else {
		    # _no_ absolute time intervals defined for the first OB
		    set hasAbsoluteTime false
		}

		# evmInfo "StartTime: $earliestStartTime  EndTime: $earliestEndTime"
	    } else {
		# for all other OBs get the Earliest after Previous values and sum them up
		set tlc [$objRef getTimeLinkConstraints $i]
		set eap [$tlc getEarliestAfterPreviousInHours]

		#also get latest after previous
		set lap [$tlc getLatestAfterPreviousInHours]
		#If latest_after_previous - earliest_after_previous < 3600sec, issue an ERROR/WAIVER:
		if { $lap != 0 && [expr $lap - $eap] < 1 } {
		    set msg "$containerName ($obName): The time interval between the earliest after and latest after the previous observation cannot be < 1hour."
		    issueErrorWaiver $containerID $msg $runID 1
		}

		set sumEaP [expr $sumEaP + $eap]
		#evmInfo "sumEaP: $sumEaP "
	    }
	}
	# TLvalue = [T2 - T1] + sum(EaP) over all OBs
	set TLvalue [expr ($earliestEndTime/3600) - ($earliestStartTime/3600) + $sumEaP]
	#evmInfo "TLvalue: $TLvalue"

	if { $obRunMode == "s" && $hasAbsoluteTime } {
	    # 6 months = 6 x 30d x 24h = 4320h
	    if { $TLvalue > 4320 } {
		issueWarning $containerID "$containerName: Time delay is larger than 6 months."
	    }
	} else {
	    # 5 months = 5 x 30d x 24h = 4320h
	    if { $obRunMode == "s" && $TLvalue > 3600 } {
		issueWarning $containerID "$containerName: Time delay is larger than 5 months."
	    }
	}

    }
    #END Sanity check for time delays in TIME-LINK containers
    #Check if concatenation or a time-link contains only a single OB [General Check]

    if { $obRunMode == "s" && $objSelected } {
	if { ( $objType == "TIME_LINK" || $objType == "CONCATENATION") &&  \
	    [llength $listOfObsBlocks] == 1 } {
	    set msg "$containerName: only has a single OB. Time-links and Concatenations should include more than one OB."
	    issueError $containerID $msg $runID
	}
    }

    #
    # Loop on all OBs
    #
    # list of OB absolute time interval lists
    set concat_OB_absTI_list {}

    #vars for weather constraints:
    set concat_seeing_list {}
    set concat_fli_list {}
    set concat_transp_list {}
    set concat_twilight_list {}
    set concat_pwv_list {}

    #list of ra,dec pairs
    set ra_dec_list {}

    set totalTimeConcat 0
    set obIdx 0
    #seconds in a standard year (365.242199 x 86,400)
    set secsInYear 31556926

    # vars for LST checks
    set concat_LST_list {}
    set hasLSTinConcat 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	set ob_id [lindex $ob 1]
	set obRunMode $OBdata($ob_id,runMode)
	set objType [lindex $ob 0]
	set objRef [lindex $ob 2]

	# Get OB name, run ID and run mode
	set obName $OBdata($ob_id,name)

	#
	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	set execTime [evmExecutionTime $objRef]

	# Compute the total execution time of selected OBs/containers
	# (use global variable for use outside of main EVM loop)
	set OBdata(totalExecutionTime) [expr $OBdata(totalExecutionTime) + $execTime]


	# collect exectution times and weather conditions/constraints:
	set csSeeing  $OBdata($ob_id,csSeeing)
	set csFli $OBdata($ob_id,csFLI)
	set csTransp $OBdata($ob_id,csTransp)
	set csTwilight $OBdata($ob_id,csTwilight)
	set csPWV $OBdata($ob_id,csPWV)


	# collect ra,dec pairs
	set ra [expr $OBdata($ob_id,ra_marcsec) / 1000.]
	set dec [expr $OBdata($ob_id,dec_marcsec) / 1000.]
	set ra_dec [list $ra $dec]


	# vv SETUP FOR LATER CONTAINER-LEVEL CHECKS vv

	if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
	    # add to total exectime for a check (outside foreach loop)
	    set totalTimeConcat [expr $totalTimeConcat + [evmExecutionTime $objRef]]

	    set tis [evmTimeIntervals $objRef]
	    #create a list of OB {startTime endTime} pairs
	    set obTimeIntervalsList {}
	    if { [llength $tis] } {
		foreach ti $tis {
		    set startTime [evmStartTime $ti]
		    set endTime [evmEndTime $ti]
		    lappend obTimeIntervalsList [list $startTime $endTime]
		}
	    }

	    #if any absolute time intervals are defined, then append the sorted list to the absolute time intervals list of the
	    #concatenation
	    if { [llength $obTimeIntervalsList] != 0} {
		#sort this list of time intervals since they may not be in chronological order (use custom sort obTIsort)
		set obTimeIntervalsList [lsort -command obTIsort $obTimeIntervalsList]
		# collect OB absolute time interval lists
		lappend concat_OB_absTI_list $obTimeIntervalsList
	    }

	    # collect WEATHER CONSTRAINTS
	    lappend concat_seeing_list $csSeeing
	    lappend concat_fli_list $csFli
	    lappend concat_transp_list $csTransp
	    lappend concat_twilight_list $csTwilight
	    lappend concat_pwv_list $csPWV

	    #collect ra,dec pairs
	    lappend ra_dec_list $ra_dec
	}


	if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
	    if { $OBdata($ob_id,LST,nbIntervals) > 0 } {
		set hasLSTinConcat 1
	    }
	}

	set ob_LSTlist {}

	for {set ti 1} {$ti <= $OBdata($ob_id,LST,nbIntervals)} {incr ti} {

	    set lst_min [expr $OBdata($ob_id,LST,$ti,start)/3600.] ; # hours
	    set lst_max [expr $OBdata($ob_id,LST,$ti,end)/3600.] ; # hours

	    #collect this here for use later in the LST checks for concats
	    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
		#collect all the time intervals for this OB , then see end of for loop
		lappend ob_LSTlist [list $lst_min $lst_max]
	    }

	} ; #end for

	# sort and add the intervals for each OB along with the OBidx; now see code at the end of OB loop (LST check for concat)
	lappend concat_LST_list [list $obIdx [lsort -command obTIsort $ob_LSTlist]]


	# ^^ END OF SETUP FOR LATER CONTAINER-LEVEL CHECKS ^^


	# vv Checks below are for OB-level only in this loop vv

	# Is the OB actually part of the selection? If not, skip the OB-level checks...
	if { !$isOBSelected } { continue }

	# Check OB name for unallowed characters:
	if { [regexp {^[a-zA-Z0-9+_ \.-]+$} $obName] == 0 } {

	    set msg "$ob_id: OB with forbidden characters in its name, please fix it.\n      (allowed characters: letters, digits, space, + - _ . )"
	    issueError $ob_id $msg $runID

	}

	# [General Check] Check OB name for "No name":
	if { [regexp -nocase {No[_\. -]*name} $obName] != 0 } {

	    set msg "$ob_id: Please provide an OB name which does not contain 'No name'."
	    issueError $ob_id $msg $runID

	}

	# [General Check] ToO/RRM prefix:
	if { $Phase1data(tooFlag) == 1 } {

	    if { ![regexp -nocase {^(TOO_|RRM_)} $OBdata($ob_id,name)] } {
		set msg "This is a ToO/RRM run, please add the prefix 'TOO_' or 'RRM_' to the name of the OB\n  depending on the kind of OB."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }

	} else {

	    if { [regexp -nocase {^(TOO_)} $OBdata($ob_id,name)] } {
		set msg "This is not a ToO run, please remove the prefix 'TOO_' from OB name."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    } 
	    if { [regexp -nocase {^(RRM_)} $OBdata($ob_id,name)] } {
		set msg "This is not a ToO/RRM run, please remove the prefix 'RRM_' from OB name."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    } 

	}

	# Check OB name - naming rule for GTO runs
	if { ![regexp {^177} $runID] && ![regexp {^([Tt][oO][oO]_)?(GTO_INAF|GTO_OCAM)} $obName] } {

	    set msg "If your run is part of a GTO programme your OB names must start with GTO_INAF or GTO_OCAM"
	    issueWarning $ob_id $msg

	}

	#
	#  [General Check] Check execution time of OB
	#
	# (execTime has been set at the beginning of the OB loop)


	if { $execTime > 3900 } {

	    set printedTime [formatPrintedTime $execTime]

	    set msg "Execution time is greater than 1h05min (found $printedTime)"
	    if { $obRunMode == "s" } {
		issueErrorWaiver $ob_id $msg $runID 3
	    } else {
		issueWarning $ob_id $msg
	    }
	}

	# [General Check] Check target name, coordinates and airmass constraints including zenith avoidance
	set targetOK [checkTarget $ob_id $runID]

	#
	# [General Check] Time Constraints Check
	#
	#Get the time intervals set
	set tis [evmTimeIntervals $objRef]
	if { [llength $tis] } {

	    #Check if ToO/RRM (but not TOO_MOV_) OB has absolute time intervals, and exit if error
	    if { $Phase1data(tooFlag) == 1 && $obRunMode == "s" && ![regexp -nocase {^(TOO_|RRM_)?MOV_} $OBdata($ob_id,name)] } {
		set msg "Please remove absolute time windows because this is a TOO/RRM observation. \n"
		append msg "        The time constraint for execution is provided when the trigger is issued."

		issueError $ob_id $msg $runID

		continue
	    }

	    set tt 0
	    set currentTime [clock seconds]
	    set earliestStartTime $currentTime
	    set latestEndTime $currentTime
	    # in case the returned set of time intervals in not ordered, get the earliest
	    foreach ti $tis {
		set startTime [evmStartTime $ti]
		set endTime [evmEndTime $ti]
		set duration [expr $endTime-$startTime]

		# Zero time interval [General Check]
		if { $duration == 0 } {
		    set msg "OB cannot have a defined absolute time interval of zero hours."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		} elseif { $duration <= 1800. } {
		    # Short time interval [General Check]
		    set msg "An absolute time interval for this OB is less than 30 minutes. Please\n"
		    set msg "$msg consider extending it in order to increase the chances for this OB to be scheduled."
		    issueWarning $ob_id $msg
		} elseif { $duration > $secsInYear } {
		    # Long time interval [General Check]
		    set msg "The OB cannot have an absolute time interval longer than 1 year."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		}
		# For Expired time intervals [General Check]
		if { $tt == 0 } {
		    set earliestStartTime $startTime
		    set latestEndTime $endTime
		} else {
		    if { $startTime < $earliestStartTime } {
			set earliestStartTime $startTime
		    }
		    if { $endTime > $latestEndTime } {
			set latestEndTime $endTime
		    }
		}
		incr tt
	    }
	    # - If the minimum date/time is in the past (before now)
	    if { $latestEndTime <= $currentTime } {
		set msg "Absolute time intervals must refer to some time in the future."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    } elseif { $earliestStartTime <= $currentTime } {
		# - If the minimum date/time is in the past (before now)
		set msg "The earliest start date/time defined in the absolute time interval(s) has already passed."
		issueWarning $ob_id $msg
	    }
	}

	# [General Check] Check target for Zenith avoidance and airmass constraint:
	set targetGoesCloseToZenith 0
	if { $targetOK } {

	    # Target info
	    set targDelta [expr $OBdata($ob_id,dec_marcsec) / 3600000.]
	    set csAirmass $OBdata($ob_id,csAirmass)


	    #the airmass at z=5 is (1./cos(5./$DegreePerRadian)) = 1.00381983748
	    if { $csAirmass < 1.004 } {				

		issueError $ob_id "Airmass cannot be set to $csAirmass." $runID

	    } else {

		set csAirmass $OBdata($ob_id,csAirmass)

		# Latitude of VLT Observatory (Paranal = -24:37:30)
		set latitude -24.63

		# Conversion factor rad => deg
		set DegreePerRadian 57.29578

		# Value of PI
		set PI 3.14159

		# Get sin(latitude) * sin(targDelta/3600.)
		set AA [expr sin($latitude/$DegreePerRadian) * sin($targDelta/$DegreePerRadian)]

		#Get cos(latitude) * cos(targDelta/3600.)
		set BB [expr cos($latitude/$DegreePerRadian) * cos($targDelta/$DegreePerRadian)]

		#Get the equivalent altitude (in radians) for the airmass constraint sin(asin(1/airmass)) = 1/airmass
		set altX_csAirmass [expr 1.0/$csAirmass]

		if { -31.63 <= $targDelta && $targDelta <= -17.63 } {

		    #Get the hour angle of the target at the zenith avoidance altitude of 83deg sin(83deg)
		    set altX_zenithAvoidance [expr sin(83./$DegreePerRadian) ]

		    set HAz [expr (24./360.) * acos( ( $altX_zenithAvoidance - $AA) / $BB ) * $DegreePerRadian]

		    #Get the hour angle of the target when the target rises above the altitude of the airmas constraint
		    set HAa [expr (24./360.) * acos( ( $altX_csAirmass - $AA) / $BB ) * $DegreePerRadian]

		    set exec_hours [expr $execTime / 3600. ]
		    if { ($HAa - $HAz) < $exec_hours } {
			set printedTime [formatPrintedTime $execTime]
			set pHAa [formatPrintedTime [expr $HAa * 3600]]
			set pHAz [formatPrintedTime [expr $HAz * 3600]]
			set msg "The execution time of $printedTime is greater than the time available for the observation.\n     HA at which target rises above the airmass constraint ($csAirmass): $pHAa \n     HA at which target reaches zenith avoidance: $pHAz "
			issueError $ob_id $msg $runID
		    } else {
			if { $obRunMode == "s" } {
			    issueWarning $ob_id "The target crosses the VST zenith target avoidance region.\n         You may wish to review your airmass constraint to avoid interrupted observations."	
			} else {
			    issueWarning $ob_id "The target crosses the VST zenith target avoidance region. Please take this\n       into account when scheduling the observations."
			}
			set targetGoesCloseToZenith 1
		    }

		} else {
		    set dec [expr $OBdata($ob_id,dec_marcsec) / 3600000. ]
		    set zenthDistance [expr abs($latitude - $dec)]
		    #get min airmass
		    set targAirmassMin [expr (1. / cos($zenthDistance/ $DegreePerRadian) ) ]
		    #get max Airmass
		    set zenthDistanceMax [expr (abs(-90 - $dec) * 2 + $zenthDistance)]
		    #this is needed for circumpolar stars
		    set targAirmassMax [expr (1. / cos($zenthDistanceMax/ $DegreePerRadian) ) ]

		    # only if the targAirmassMax > csAirmass i.e. the target gets below the airmass constraint at some point and can rise above it
		    # do the following check; Especially required for circumpolar objects. Note $targAirmassMax is -ve for +dec objects

		    set refDec [expr -90 - $dec]

		    if { $dec >= $refDec || ( $dec < $refDec && $targAirmassMax > $csAirmass)  } {

			#Get the hour angle of the target when the target rises above the altitude of the airmas constraint
			set HAa [expr (24./360.) * acos( ( $altX_csAirmass - $AA) / $BB ) * $DegreePerRadian]

			set exec_hours [expr $execTime / 3600. ]
			set tavail [expr $HAa * 2]
			set exec_hours [expr $execTime / 3600. ]
			set printedTime [formatPrintedTime $execTime]
			set pHAa [formatPrintedTime [expr $HAa * 3600]]
			set ptA [formatPrintedTime [expr $tavail * 3600]]

			if { ( $tavail ) < $exec_hours && $obRunMode == "s" } {
			    set printedTime [formatPrintedTime $execTime]
			    set pHAa [formatPrintedTime [expr $HAa * 3600]]
			    set ptA [formatPrintedTime [expr $tavail * 3600]]
			    set msg "The execution time of $printedTime is greater than the time available for the observation, given your airmass constraint.\n     HA at which target rises above the airmass of $csAirmass : $pHAa \n     Time available for observation: $ptA "
			    issueError $ob_id $msg $runID
			}
		    }
		} ; #End else - dec not in zenith avoidence zone
	    } ; #End the airmass at z=5
	} ;# End Check target for Zenith avoidance and airmass constraint


	# [General Check] Check epoch in range 1950.0-2050.0
	checkEpoch $ob_id $runID    

	# [General Check] check on proper motions
	checkProperMotions $ob_id


	# [General Check] TOO/RRM in container check
	if { $obRunMode == "s" } {
	    if { $Phase1data(tooFlag) == 1 && $isContainer } {

		if { [regexp -nocase {^(TOO_)} $OBdata($ob_id,name)]  && $containerType == "TIME_LINK"} {
		    set msg "TOO OBs should not be placed in a Time link."
		    issueError $ob_id $msg $runID
		}
		if { [regexp -nocase {^(RRM_)} $OBdata($ob_id,name)] } {
		    set msg "RRM OBs should not belong in a Group, Concatenation, or Time link. "
		    issueError $ob_id $msg $runID
		}
	    }
	}

	# PWV Constraint [General Check]. Value for default must be the same as in the constraints xsd files
	checkPWV $ob_id 30

	#
	# Check for possible vignetting by wind mast: this occurs for elevations < 45deg and
	# azimuth between 157 and 167.
	# We only make the check if the airmass constraint allows one to observe below 45deg

	if { $OBdata($ob_id,csAirmass) >= 1.4 } {

	    # Define the sky area where the vignetting by the wind mast occurs:

	    # min and max elevation
	    set minAlt 15
	    set maxAlt 45

	    # min and max azimuth
	    set minAz 157
	    set maxAz 167

	    # Conversion factors degrees <--> radians
	    set deg2rad [expr 3.14159265358979 / 180.]
	    set rad2deg [expr 180. / 3.14159265358979]

	    # Paranal latitude
	    set lat -24.63

	    # Target coordinates

	    set alpha [expr $OBdata($ob_id,ra_marcsec) / (15 * 3600000.)] ; # hours
	    set delta [expr $OBdata($ob_id,dec_marcsec) / 3600000.] ; # degrees

	    # internal flag used for recording of LST interval
	    set inFlag false
	    set vignetting false

	    # Some expressions needed later for Alt-Az <--> HA-Dec conversion
	    # sin(lat)*sin(delta)
	    set a [expr sin($deg2rad*$lat)*sin($deg2rad*$delta)]
	    # cos(lat)*cos(delta)
	    set b [expr cos($deg2rad*$lat)*cos($deg2rad*$delta)]

	    # Loop on LST
	    for {set LST 0} {$LST < 24} {set LST [expr $LST + 1/6.]} {

		# Hour angle in hours
		set HA [expr $LST - $alpha]

		# Hour angle in degrees
		set HA_deg [expr 15 * $HA]

		# Target altitude in degrees
		set sinh [expr $a + $b * cos($deg2rad*$HA_deg) ]
		set h [expr asin($sinh)/$deg2rad]

		# Target azimuth in degrees (counted from South, positive towards West):
		# one adds 180 degrees to normal definition of azimuth, to
		# cope with ESO convention
		set x [expr sin($deg2rad*$HA_deg) * cos($deg2rad*$delta) * cos($deg2rad*$lat)]
		set y [expr -1.* ($sinh * sin($deg2rad*$lat) - sin($deg2rad*$delta))]
		set azim [expr 180. - atan2($x,$y) * $rad2deg]

		if { $azim >= 360.0 } { set azim [expr $azim - 360.] }

		# If the target is within the vignetting area
		if { $minAz <= $azim && $azim <= $maxAz && $minAlt <= $h && $h <= $maxAlt } {

		    set vignetting true

		    # If this is the first point we find, record the beginning of the LST interval
		    if {$inFlag == false} {
			set vignetting_start $previousLST
			set inFlag true
		    }

		} elseif { $inFlag == true } {

		    # Record end of LST interval
		    set vignetting_end $LST
		    set inFlag false

		}

		set previousLST $LST
	    }

	    if { $vignetting == true } {

		# LST in printable format hh:mm
		set LSTstart [format "%02d:%02d" [expr int($vignetting_start)] [expr int(60*($vignetting_start - int($vignetting_start)))]]
		set LSTend [format "%02d:%02d" [expr int($vignetting_end)] [expr int(60*($vignetting_end - int($vignetting_end)))]]

		if { $OBdata($ob_id,LST,nbIntervals) == 0 } {

		    # LST in printable format hh:mm
		    set LSTstart [format "%02d:%02d" [expr int($vignetting_start)] [expr int(60*($vignetting_start - int($vignetting_start)))]]
		    set LSTend [format "%02d:%02d" [expr int($vignetting_end)] [expr int(60*($vignetting_end - int($vignetting_end)))]]

		    set msg "Your target may pass by the area with wind mast vignetting, which happens between azimuth angles 157-167 at altitudes < 45 degrees.\n         To avoid vignetting, the LST constraints should be set such that LST \[$LSTstart - $LSTend\] is avoided."
		    issueWarning $ob_id $msg

		} else {

		    set vstart $vignetting_start
		    set vend $vignetting_end
		    if { $vend < $vstart } { set vend [expr 24 + $vend] }

		    # Prepare list of LST intervals (min and max in hours)
		    for {set i 1} {$i <= $OBdata($ob_id,LST,nbIntervals) } {incr i} {

			set start $OBdata($ob_id,LST,$i,start)
			set start [expr ($start/3600.) - ($start/3600)/24 * 24]
			set lst_min $start

			set end $OBdata($ob_id,LST,$i,end)
			set end [expr ($end/3600.) - ($end/3600)/24 * 24]
			set lst_max $end

			if { $end < $start } { set end [expr 24 + $end] }

			# Overlapping intervals?
			if { $start < $vend && $vstart < $end } {

			    set lst_Min_hrs [expr int($lst_min)]
			    set lst_Min_min [expr round(($lst_min - int($lst_min))*60)]
			    set lst_Min [format "%02d:%02d" $lst_Min_hrs $lst_Min_min]
			    set lst_Max_hrs [expr int($lst_max)]
			    set lst_Max_min [expr round(($lst_max - int($lst_max))*60)]
			    set lst_Max [format "%02d:%02d" $lst_Max_hrs $lst_Max_min]

			    set msg "For the selected LST interval \[$lst_Min - $lst_Max\], your target will pass by the area with wind mast vignetting, which happens between azimuth angles 157-167 at altitudes < 45 degrees.\n         To avoid vignetting, the LST constraints should be set such that LST \[$LSTstart - $LSTend\] is avoided."
			    issueWarning $ob_id $msg

			}
		    }

		}
	    }
	}

	# Check for over-specified seeing constraint
	if { $OBdata($ob_id,type) != "C"  && $OBdata($ob_id,csSeeing) < 0.6 } {

	    set msg "Seeing constraint is smaller than 0.6\""
	    issueWarning $ob_id $msg $runID
	}

	if { $Phase1data(isPublicSurvey) != 1 } {

	    #######################################
	    # Phase 2 vs. Phase 1 compliance checks (SM only)
	    #######################################

	    if { $obRunMode == "s" } { 
		## NOTE 25Jun2015 SMA: to avoid issues with some ongoing large programmes we
		# still perform the 'old' EVM check on seeing for the concerned programmes:
		# 177.A-3016 177.D-3023
		if { [lsearch -exact {177.A-3016 177.D-3023} [string range $runID 0 end-3] ] != -1 } {

		    # Seeing constraint

		    if { $OBdata($ob_id,csSeeing) < $Phase1data(seeing) } {

			set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The seeing value allowed for this OB is >= $Phase1data(seeing) arcsec."
			issueError $ob_id $msg $runID

		    }

		    if { $OBdata($ob_id,csSeeing) > $Phase1data(seeing) } {

			set msg "The Seeing constraint for this OB has been relaxed with respect to that in the Phase 1 proposal ($Phase1data(seeing) arcsec). Is this intentional?"
			issueInfo $ob_id $msg

		    }

		} else {
		    # Check expected image quality
		    checkImageQuality $objRef $runID 
		}


		# Transparency constraint

		if { $OBdata($ob_id,csTransp) == "1PHO" && $Phase1data(skyTransp) == "CLR" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Sky Transparency constraint for this OB is 'Clear', 'Thin' or 'Thick'."
		    issueWarning $ob_id $msg

		} elseif { ($OBdata($ob_id,csTransp) == "1PHO" || $OBdata($ob_id,csTransp) == "2CLR") \
			&& $Phase1data(skyTransp) == "THN" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Sky Transparency constraint for this OB is 'Thin' or 'Thick'."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } 

		} elseif { ($OBdata($ob_id,csTransp) == "2CLR" || $OBdata($ob_id,csTransp) == "3THN" || $OBdata($ob_id,csTransp) == "4THK") \
			&& $Phase1data(skyTransp) == "PHO" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { ($OBdata($ob_id,csTransp) == "3THN" || $OBdata($ob_id,csTransp) == "4THK") && \
			$Phase1data(skyTransp) == "CLR" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { $OBdata($ob_id,csTransp) == "4THK" && $Phase1data(skyTransp) == "THN" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		}

		# Moon FLI constraint

		if { $OBdata($ob_id,csFLI) < 0.4 && $Phase1data(moonFLI) == "g" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n            The allowed Lunar Illumination value for this OB is: 0.4 <= FLI <= 0.7 (grey) or FLI > 0.7 (bright)."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } 

		} elseif { $OBdata($ob_id,csFLI) < 0.4 && $Phase1data(moonFLI) == "n" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Lunar Illumination value for this OB is: FLI > 0.7"
		    issueWarning $ob_id $msg

		} elseif { $OBdata($ob_id,csFLI) >= 0.4 && $OBdata($ob_id,csFLI) <= 0.7 && $Phase1data(moonFLI) == "n" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n            The allowed Lunar Illumination value for this OB is: FLI > 0.7"
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } 

		} elseif { $OBdata($ob_id,csFLI) >= 0.4 && $Phase1data(moonFLI) == "d" } {

		    set msg "The Lunar Illumination constraint for this OB has been relaxed with respect to that in Phase 1. Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { $OBdata($ob_id,csFLI) > 0.7 && $Phase1data(moonFLI) == "g" } {

		    set msg "The Lunar Illumination constraint for this OB has been relaxed with respect to that in Phase 1. Is this intentional?"
		    issueInfo $ob_id $msg

		}

		# Checks on Phase 1 targets

		# Find closest Phase 1 target
		set closest_ti 0
		set closest_dist 1e99

		for { set ti 1 } { $ti <= $Phase1data(target,nbTargets) } { incr ti } {

		    # Compute target distance
		    set distance [computeDistance \
			$Phase1data(target,$ti,RA_arcsec) [expr $OBdata($ob_id,ra_marcsec)/1000.] \
			$Phase1data(target,$ti,DEC_arcsec) [expr $OBdata($ob_id,dec_marcsec)/1000.]  ]

		    if { $distance < $closest_dist } {
			set closest_dist $distance
			set closest_ti $ti
		    }
		}

		# Set max allowed distance to original target in arcsec
		set targetRadius 2700

		if { $Phase1data(tooFlag) != 1 && \
		    [regexp -nocase {^(MOV_|CAL_|PSF_)} $OBdata($ob_id,name)] == 0 && \
			$closest_dist > $targetRadius } {

		    set msg "The target seems to be missing from the target list in the proposal."
		    issueWarning $ob_id $msg

		}

		if { $Phase1data(tooFlag) != 1 && \
		    [regexp -nocase {^(MOV_|CAL_|PSF_)} $OBdata($ob_id,name)] == 0 && \
			$closest_dist <= $targetRadius && \
		    [string first $runLetter $Phase1data(target,$closest_ti,runRefs)] == -1} {

		    set msg "This target seems to be missing from the target list for this run of the proposal."
		    issueWarning $ob_id $msg

		}
	    }; # End of SM Only Phase2/Phase1 compliance checks

	    # Checks on LST constraints [General Check]
	    #  OB duration must be less than LST range(s)

	    for {set ti 1} {$ti <= $OBdata($ob_id,LST,nbIntervals)} {incr ti} {

		set LSTduration [expr $OBdata($ob_id,LST,$ti,end) - $OBdata($ob_id,LST,$ti,start)] ; # seconds

		if { $LSTduration < $execTime } {
		    set msg "One or more LST ranges are shorter than the execution time of the OB."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		    break
		}
	    } ; #end for

	}; # End is Public Survey

	#
	# Optional: checks on attached ephemeris file (SM only)
	#
	if { $obRunMode == "s" } {
	    # Retrieve ephemeris file + data
	    set ephFile [evmEphemerisFile $objRef]
	    set ephFileData [evmEFData $ephFile]

	    # Check consistency of OB name if an ephemeris file is present:
	    # must have the prefix 'MOV_'
	    if { $ephFileData != "" && ![regexp -nocase {^(TOO_|RRM_)?MOV_} $OBdata($ob_id,name)] } {
		set msg "When using an ephemeris file please prefix the name of the OB with 'MOV_' or 'TOO_MOV_' or 'RRM_MOV_' as the case may be."
		issueError $ob_id $msg $runID
	    }

	    if { $ephFileData == "" && [regexp -nocase  {^(TOO_|RRM_)?MOV_} $OBdata($ob_id,name)] } {
		set msg "The prefix 'MOV_' in the OB name suggests a moving target but no ephemeris data was found."
		issueError $ob_id $msg $runID
	    }

	    # Check that each OB with an ephemeris file has an absolute time constraint
	    set ti [evmTimeIntervals $objRef]
	    if { $ephFileData != "" && [llength $ti] == 0 } {
		set msg "When using an ephemeris file the OB must have some absolute time constraint defined."
		issueError $ob_id $msg $runID
	    }

	    #Check if no ephemeris and warning if absolute time intervals [General Check]
	    if { [string trim $ephFileData] == "" && [llength $ti] != 0  } {				
		set msg "If absolute time windows have been specified only to ensure compliance with the \n"     
		append msg "       airmass constraint, then please remove the time windows as they are unnecessary.\n" 
		append msg "       But, if you choose not to remove them, then please justify the absolute time windows \n" 
		append msg "       in the 'Readme' file .\n"
		issueWarning $ob_id $msg
	    }

	    # Check contents of ephemeris file (separate procedure)
	    if { $ephFileData != "" } { checkEphemerisFile $ob_id $runID $ephFileData $ti }

	}; #End of SM only Ephemeris related checks


	# Check on FC metadata [General Check]
	checkFC_Metadata $ob_id $objRef

    }; # end of loop on OBs

    # ----------------------
    # CONTAINER-LEVEL CHECKS
    # ----------------------

    # [General Check] Check the CONSISTENCY OF THE WEATHER CONSTRAINTS within a concatenation
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {
	if { [llength [uniq $concat_seeing_list]] > 1 } {
	    issueWarning $containerID "$containerName: The seeing constraint is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_fli_list]] > 1 } {
	    issueWarning $containerID "$containerName: The FLI is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_transp_list]] > 1 } {
	    issueWarning $containerID "$containerName: The transparency is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_twilight_list]] > 1 } {
	    issueWarning $containerID "$containerName: The twilight is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_pwv_list]] > 1 } {
	    issueWarning $containerID "$containerName: The PWV is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
    }

    # [General Check] Compute RA Range of targets within concatenation to ensure feasibility
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {

	set lim75 [format "%.2f" [expr 75 * 3600]]
	set lim45 [format "%.2f" [expr 45 * 3600]]
	set maxSep [angularSeparationCheck $ra_dec_list]

	if { $maxSep > $lim75 } {

	    set msg "Due to the large angular separation between the targets (>75 deg), the scheduling of such a \n container is extremely difficult."
	    issueErrorWaiver $containerID "$containerName: $msg" $runID 11

	} elseif { $maxSep > $lim45 } {

	    set msg "Due to the large angular separation between the targets (>45 deg), the scheduling of such a \n container is particularly difficult."
	    issueWarning $containerID "$containerName: $msg"
	}
    }

    #  [General Check] Total duration of concatenation must not exceed 65 minutes
    if { $isContainer && $containerType == "CONCATENATION" && $objSelected && \
	    $totalTimeConcat > 3900 && $obRunMode == "s" } {

	set printedTime [formatPrintedTime $totalTimeConcat]

	set msg "$containerName: Execution time for this concatenation exceeds 1h05m (found $printedTime)"
	issueErrorWaiver $containerID $msg $runID 2
    }

    #  [General Check] Absolute time windows of a Concatenation
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {
	if { [llength [uniq $concat_OB_absTI_list]] > 1 } {
	    set msg "$containerName: The absolute time intervals are not the same for all OBs in the container."
	    issueError $containerID $msg $runID
	}
    }

    #  LST in a Concatenation [General Check]
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected && $hasLSTinConcat } {

	set msg ""
	#puts "concat_LST_list $concat_LST_list [llength $concat_LST_list]"
	#determine each of the continuous intervals present in the set of LST intervals
	set continuousLSTlist {}
	set numOBs [llength $concat_LST_list]
	if { $numOBs > 0 } {
	    #add intervals of the first OB
	    set continuousLSTlist [lindex [lindex $concat_LST_list 0] 1]
	    #puts "continuousLSTlist $continuousLSTlist"
	    # define newViableForNextOB which will contain the list of defined intervals that the next OB will be checked against, start with the intervals of the first OB
	    # add all intervals to the list of continuous LST intervals starting with the first
	    set newViableForNextOB $continuousLSTlist

	    # for each OB (starting with the second OB) in the concatenation
	    for {set k 1} {$k < $numOBs} {incr k} {
		#get list of intervals for the OB
		set ob_LSTlist [lindex [lindex $concat_LST_list $k] 1]
		#for each interval in the continuousLSTlist, see if you need to change the bounds and then get a new list of viable intervals
		# now the new list of viable intervals becomes the continuousLSTlist before going to the next OB
		set continuousLSTlist $newViableForNextOB

		#puts "*** starting OB #[expr $k + 1] check against $continuousLSTlist"
		# reset newViableForNextOB
		set newViableForNextOB {}
		#also define newContinuousList which will contain the overlapping LST intervals which is used at the end of the concatenation
		set newContinuousList {}
		for {set jj 0} {$jj < [llength $continuousLSTlist]} {incr jj} {
		    for {set ii 0} {$ii < [llength $ob_LSTlist]} {incr ii} {
			set newInterval [unionInt [lindex [lindex $continuousLSTlist $jj] 0] [lindex [lindex $continuousLSTlist $jj] 1] \
			    [lindex [lindex $ob_LSTlist $ii] 0] [lindex [lindex $ob_LSTlist $ii] 1] ]
			#puts "newInterval $newInterval"
			if { $newInterval != [list 0 0] } {
			    lappend newContinuousList $newInterval
			    #puts "newContinuousList $newContinuousList"
			    lappend newViableForNextOB [lindex $ob_LSTlist $ii]
			}
		    } ; # end list of time intervals in the OB
		} ; # end list of time intervals in the continuousLSTlist

		if { [llength $newViableForNextOB] == 0 } {
		    set msg "$containerName: OB #[expr $k + 1] in the concatenation does not have LST intervals that overlap with OB #$k.\n        Overlap is necessary for OBs to be executed back-to-back."
		    break
		}

		# when no next OB, then newContinuousList is the continuousLSTlist else newViableForNextOB is the continuousLSTlist (beginning of the loop)
		set continuousLSTlist $newContinuousList
		#puts "viable intervals for the next OB $newViableForNextOB length [llength $newViableForNextOB]"
	    } ; # END for each OB

	    if { $msg != "" } {
		issueError $containerID $msg $runID
	    } else {
		#Now check for duration of LST
		#puts "continuousLSTlist $continuousLSTlist"
		for {set tj 0} {$tj < [llength $continuousLSTlist]} {incr tj} {
		    set startLST [lindex [lindex $continuousLSTlist $tj] 0]
		    set endLST  [lindex [lindex $continuousLSTlist $tj] 1]

		    set LSTduration [expr ($endLST - $startLST)  * 3600]
		    #puts "startLST $startLST endLST $endLST ---- LST duration [formatPrintedTime $LSTduration]"
		    if { $LSTduration < $totalTimeConcat } {
			set printedTime [formatPrintedTime $totalTimeConcat]
			set msg "$containerName: The LST interval(s) available to execute the concatenation in should be greater than the execution time of $printedTime"
			issueError $containerID $msg $runID
			break
		    }
		}
	    }
	} ; # end if num of OBs > 0
    } ; # end LST in concat check 

}; #End proc checkOMEGACAMlevel1

#### OMEGACAM Level 2 Checks - general OB structure

proc checkOMEGACAMLevel2 { object } {

    global evmOMEGACAMTpl
    global OBdata

    set templateListInstrument [getArrayElements evmOMEGACAMTpl]

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    #
    # Loop on all OBs
    #

    set obIdx 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	set ob_id [lindex $ob 1]

	# vv SETUP FOR LATER CONTAINER-LEVEL CHECKS vv

	####### nothing to setup #########

	# ^^ END OF SETUP FOR LATER CONTAINER-LEVEL CHECKS ^^

	# Is the OB actually part of the selection? If not, skip it..
	if { !$isOBSelected } { continue }

	# vv Checks below are for OB-level only in this loop vv


	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	# Get name of acquisition
	set acqName $OBdata($ob_id,0,tplName)

	set tplListOB {}
	set tplListPattern {}

	# Loop on all science templates
	for { set i 1 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    set tplName $OBdata($ob_id,$i,tplName)
	    set tplDitherPattern $OBdata($ob_id,$i,TEL.TARG.PATTERN)
	    lappend tplListOB $tplName
	    lappend tplListPattern $tplDitherPattern

	}

	set tplListOB [uniq $tplListOB]
	set tplListPattern [uniq $tplListPattern]

	if { $acqName == "OMEGACAM_img_acq_movetogap" } {

	    if {$tplListOB != "OMEGACAM_img_obs_dither" } {

		set msg "$acqName (\#0): when using the 'movetogap' acquisition one is supposed to use the 'dither' template (and 'starext' pattern) for the observations."
		issueWarning $ob_id $msg

	    } elseif {$tplListPattern != "starext"} {

		set msg "$acqName (\#0): when using the 'movetogap' acquisition one is supposed to use the 'starext' dither pattern for the observations."
		issueWarning $ob_id $msg
	    }

	}

	# Loop on all science templates
	for { set i 1 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    set tplName $OBdata($ob_id,$i,tplName)

	    if { $tplName == "OMEGACAM_img_obs_dither" && $OBdata($ob_id,$i,TEL.TARG.PATTERN) == "starext" && \
		    $acqName != "OMEGACAM_img_acq_movetogap" } {

		set msg "$tplName (\#$i): when using this template with the 'starext' dithering pattern, you should\n         indicate in the readme file and finding chart if a particular source should go into a detector gap. "
		issueWarning $ob_id $msg

	    }

	}

    }; # end of loop on OBs


    # ----------------------
    # CONTAINER-LEVEL CHECKS
    # ----------------------

    ####### no container-level checks #########

    # -----------------------------
    # END OF CONTAINER-LEVEL CHECKS
    # -----------------------------
}

#### OMEGACAM Level 3 Checks - detailed checks on parameters

proc checkOMEGACAMLevel3 { object } {

    global OBdata evmDITLimits

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]
	set containerType [lindex $containerInfo 0] ; # Type of container

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    #
    # Loop on all OBs
    #

    set obIdx 0

    set timeSinceLastIA 0

    set prevOBsLastTplFilter ""

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	set objRef [lindex $ob 2]
	set ob_id [lindex $ob 1]

	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	# Get name of acquisition
	set acqName $OBdata($ob_id,0,tplName)

	# vv CONTAINER-LEVEL CHECKS vv

	# For concatenations: checks on "Image Analysis" flag
	if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {

	    if { $OBdata($ob_id,0,OCS.IA.START) == "T" } {
		set timeSinceLastIA 0
	    }

	    set timeSinceLastIA [expr $timeSinceLastIA  + [evmExecutionTime $objRef]]
	    # What if the Image Analysis flag was set to false?
	    if { $OBdata($ob_id,0,OCS.IA.START) == "F" } {

		if { $obIdx == 1 } {

		    # flag not set, is in a concatenation and is the first OB, issue an ERROR/WAIVER (SM).
		    set msg "$acqName (OB $ob_id,\#0): Image Analysis must be set to TRUE in the first OB of any concatenation."
		    issueErrorWaiver $containerID $msg $runID 10  

		} else {

		    #flag not set, is in a concatenation and is not the first OB
		    set noerrflag 0

		    #Check 1
		    if { $timeSinceLastIA > 3600 } {
			#last Image Analysis OB is more than an hour ago, issue an ERROR (SM).
			set msg "$acqName (OB $ob_id,\#0): Image Analysis is disabled: you cannot observe more than one hour without Image Analysis during an OB concatenation."
			issueError $containerID $msg $runID

		    } else {	    
			set noerrflag 1
		    }

		    #Check 2
		    #get acquisition filter
		    set currentAcqFilter $OBdata($ob_id,0,INS.FILT.NAME)

		    if { $prevOBsLastTplFilter != "" } {
			if { $prevOBsLastTplFilter != $currentAcqFilter } {
			    # the last template of the previous OB has a different filter as the aquisition, issue an ERROR (SM)
			    set msg "$acqName (OB $ob_id,\#0): Image Anaylsis is disabled: You must enable Image Analysis\n            in the aquisition when changing the filter between OBs in concatenation."
			    issueError $containerID $msg $runID

			} else {
			    # both Check #1 and #2 should fall into their respective elses for the next warning to be printed.
			    set noerrflag $noerrflag    
			}
		    }

		    # for a pass on Checks 1 and 2 but last Image Analysis OB more than 30 minutes ago, issue an ERROR/WAIVER (SM).
		    if { $noerrflag && $timeSinceLastIA > 1800 } {    
			set msg "$acqName (OB $ob_id,\#0): you have disabled Image Analysis for more than 30 minutes within this OB concatenation."
			issueErrorWaiver $containerID $msg $runID 8

		    }

		}

	    }

	    # Store the value of the last filter in the OB...
	    set prevOBsLastTplFilter $OBdata($ob_id,[expr $OBdata($ob_id,nbTpl) - 1],INS.FILT.NAME)

	}

	# ^^ CONTAINER-LEVEL CHECKS ^^

	# Is the OB actually part of the selection? If not, skip it..
	if { !$isOBSelected } { continue }

	# vv Checks below are for OB-level only in this loop vv


	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	#
	# Checks on acquisition template
	#

	# Is the ADC being used? ERROR (SM and VM)
	if { $OBdata($ob_id,0,TEL.ADC.TYPE) != "NONE" } {

	    set msg "$acqName (\#0): The Atmospheric Dispersion Corrector (ADC) is not offered until futher notice."
	    issueError $ob_id $msg $runID
	}

	# Check the 'Image analysis flag' in SM (check for non-concatenated OBs)
	if { $obRunMode == "s" && $OBdata($ob_id,0,OCS.IA.START) == "F" && $objType != "CONCATENATION" } {

	    #flag not set and not in a concatenation, issue an ERROR/WAIVER (SM).
	    set msg "$acqName (\#0): Image Analysis must be set to TRUE for any OB that is not part of a concatenation."
	    issueErrorWaiver $ob_id $msg $runID 9

	}

	# Check the 'Image analysis flag' in VM.
	if { $obRunMode == "v" && $OBdata($ob_id,0,OCS.IA.START) == "F" } {
	    set msg "$acqName (\#0): Image Analysis should normally be set to TRUE for all OBs in Visitor Mode."
	    issueWarning $ob_id $msg 
	}

	set filters_list {}
	set exposure_list {}

	set lastTemplate [expr $OBdata($ob_id,nbTpl) - 1]
	set currentLastFilter $OBdata($ob_id,$lastTemplate,INS.FILT.NAME)

	#
	# Loop on all templates
	#

	for { set i 0 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    set tplName $OBdata($ob_id,$i,tplName)

	    # *******************************************************
	    # ==> CATEGORY: FILTERS, SLIT, GRISMS...
	    # *******************************************************

	    # Check that first science template uses same filter as acquisition
	    if { $i == 1 && $OBdata($ob_id,$i,INS.FILT.NAME) != $OBdata($ob_id,0,INS.FILT.NAME) } {

		set msg "$tplName (\#$i): the filter defined in the first science template is not identical to the one used in the acquisition."
		issueWarning $ob_id $msg

	    }

	    set csAirmass $OBdata($ob_id,csAirmass)
	    if { $i > 0 && $OBdata($ob_id,$i,INS.FILT.NAME) == "u_SDSS" && $csAirmass > 1.2 } {

		set msg "$tplName (\#$i): your airmass constraint is larger than 1.2. Please note that atmospheric dispersion may affect image quality in u_SDSS"
		issueWarning $ob_id $msg
	    }

	    set csSeeing $OBdata($ob_id,csSeeing)
	    if { $i > 0 && $OBdata($ob_id,$i,INS.FILT.NAME) == "u_SDSS" && $csSeeing < 1.0 } {

		set msg "$tplName (\#$i): Requesting FWHM < 1.0 arcsec in u-band is not recommended (low probability of realization)."
		issueWarning $ob_id $msg
	    }

	    # collect template filters + exposure times to see if there is more than one kind of filter and autoguiding is set to T.
	    # Check is outside the template loop, see below.
	    lappend filters_list $OBdata($ob_id,$i,INS.FILT.NAME)
	    if { $OBdata($ob_id,$i,DET1.WIN1.UIT1) != "" } {
		lappend exposure_list $OBdata($ob_id,$i,DET1.WIN1.UIT1)
	    }

	    if { $OBdata($ob_id,$i,INS.FILT.NAME) == "NB_659" } {

		set msg "$tplName (\#$i): The NB_659 filter is a private filter and may only be used upon prior agreement with the VPHAS+ public survey team."
		issueWarning $ob_id $msg

	    }

	    # ***************************************************
	    # ==> CATEGORY: EXPOSURE TIME AND NUMBER OF EXPOSURES
	    # ***************************************************

	    # If autoguiding is disabled check that only short exposures are used
	    if { $OBdata($ob_id,0,OCS.AG.START) == "F" && $OBdata($ob_id,$i,DET1.WIN1.UIT1) > 60 } {

		set msg "$tplName (\#$i): you have de-activated autoguiding in the acquisition but some of your exposures are larger than 60 seconds."

		if { $obRunMode == "s" } {
		    issueErrorWaiver $ob_id $msg $runID 4
		} else {
		    issueWarning $ob_id $msg
		}

	    }

	    # For the 'jitter' template check that the number of exposures is larger than 1
	    if { $tplName == "OMEGACAM_img_obs_jitter" && $OBdata($ob_id,$i,SEQ.NEXPO) <= 1 } {

		set msg "$tplName (\#$i): the number of exposures is set to 1: only one image, no jitter."
		issueWarning $ob_id $msg
	    }

	    # *******************************
	    # ==> CATEGORY: OFFSET PARAMETERS
	    # *******************************


	    # Check for non-default dithering pattern
	    if { $tplName == "OMEGACAM_img_obs_dither" } {

		if { $OBdata($ob_id,$i,SEQ.NEXPO) != 5 } {		    
		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): you are not using the recommended default values for the exposure parameter (number of offset exposures)."
		    issueWarning $ob_id $msg		    
		} 

		if { $OBdata($ob_id,$i,TEL.TARG.OFFSETSIZEX) != 25 \
			|| $OBdata($ob_id,$i,TEL.TARG.OFFSETSIZEY) != 85 } {
		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): you are not using the recommended default values for the dithering parameters (offset step sizes)."
		    issueWarning $ob_id $msg
		}

	    }

	    # Check for non-default gap sizes with the dither template
	    if { $tplName == "OMEGACAM_img_obs_dither" && $OBdata($ob_id,$i,TEL.TARG.PATTERN) == "starext" && \
		    ( $OBdata($ob_id,$i,TEL.TARG.DX) != 8 || $OBdata($ob_id,$i,TEL.TARG.DY) != 8 ) } {

		set msg "$tplName (\#$i): you are not using the default values for the x and y gaps to be left unexposed."
		issueWarning $ob_id $msg
	    }

	    # Check for too large gap sizes with the dither template
	    if { $tplName == "OMEGACAM_img_obs_dither" && $OBdata($ob_id,$i,TEL.TARG.PATTERN) == "starext" && \
		    ( $OBdata($ob_id,$i,TEL.TARG.DX) > $OBdata($ob_id,$i,TEL.TARG.OFFSETSIZEX) || \
		    $OBdata($ob_id,$i,TEL.TARG.DY) > $OBdata($ob_id,$i,TEL.TARG.OFFSETSIZEY) ) } {

		set msg "$tplName (\#$i): you are using values larger than the respective offsets for the \n                      x and/or y gaps thereby exposing any object in the gap."

		if { $obRunMode == "s" } {
		    issueErrorWaiver $ob_id $msg $runID 5
		} else {
		    issueWarning $ob_id $msg
		}

	    }

	    # Check for large cumulative offsets
	    if { $tplName == "OMEGACAM_img_obs_offset" } {

		set cumulatedOffset_X 0
		set cumulatedOffset_Y 0
		set idx_x -1
		set idx_y -1

		for {set k 1} {$k <= $OBdata($ob_id,$i,SEQ.NEXPO)} {incr k} {

		    incr idx_x
		    if { $idx_x >= [llength $OBdata($ob_id,$i,TEL.TARG.OFFSETALPHA)] } { set idx_x 0 }
		    incr idx_y
		    if { $idx_y >= [llength $OBdata($ob_id,$i,TEL.TARG.OFFSETDELTA)] } { set idx_y 0 }

		    set offsetx [lindex $OBdata($ob_id,$i,TEL.TARG.OFFSETALPHA) $idx_x]
		    set offsety [lindex $OBdata($ob_id,$i,TEL.TARG.OFFSETDELTA) $idx_y]

		    set cumulatedOffset_X [expr $cumulatedOffset_X + $offsetx]
		    set cumulatedOffset_Y [expr $cumulatedOffset_Y + $offsety]
		}

		if { $OBdata($ob_id,$i,SEQ.RET2ORIG) == "T" && \
			([expr abs($cumulatedOffset_X)] > 3600 || [expr abs($cumulatedOffset_Y)] > 3600) } {
		    # ERROR (SM and VM)
		    set msg "$tplName (\#$i): 'return to origin' with acccumulated offsets of more than 3600 arcsec along RA or DEC is not possible."
		    issueError $ob_id $msg $runID

		} elseif { [expr abs($cumulatedOffset_X)] > 1800 || [expr abs($cumulatedOffset_Y)] > 1800 } {

		    set msg [format \
			"$tplName (\#$i): the cumulated offsets in RA or DEC exceed 1800 arcseconds! (RA=%.0f arcsec / DEC=%.0f arcsec)" \
			[expr abs($cumulatedOffset_X)] [expr abs($cumulatedOffset_Y)] ]
		    #removed for P92: issueErrorWaiver $ob_id $msg $runID 6
		    issueWarning $ob_id $msg

		}
	    }

	    # Maximum jitter size
	    if { $tplName == "OMEGACAM_img_obs_jitter" && $OBdata($ob_id,0,OCS.AG.START) == "T" && \
		[expr ( $OBdata($ob_id,$i,SEQ.NEXPO) - 1 ) * $OBdata($ob_id,$i,TEL.TARG.OFFSETSIZE) ] > 100 } {

		set msg "$tplName (\#$i): large jitters: guide star may be lost."

		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}

	    }

	    # *******************************
	    # ==> CATEGORY: OTHER PARAMETERS
	    # *******************************

	}; # end of loop on templates


	# if there is more than 1 kind of filter in the ob templates, and autoguiding in acquisition is T, i.e

	set min_exposure [lindex [lsort -real -increasing $exposure_list] 0]
	regsub -all {g_SDSS|r_SDSS|i_SDSS} [uniq $filters_list] "" filters_list_filtered

	if { [llength [uniq $filters_list]] > 1 && $OBdata($ob_id,0,OCS.AG.START) == "T" && \
		$min_exposure < 120 && [llength $filters_list_filtered] != 0 } {

	    # there is more than one type of filter and autoguiding is true, some exposure times
	    # are shorter than 2 min, and some of the filters are not g r'or i', so...

	    set msg "Changing filters while autoguiding. For short exposures <~ 1-2 min some image elongation may appear."
	    issueWarning $ob_id $msg

	}

    }; # end of loop on OBs

}; # End proc checkOMEGACAMlevel3


