# File: exec_AAA_Common.tcl
# Purpose: common procedures for all execution time scripts
#
# Author: Stephane Marteau
#

# 2016-06-19: CCB-1084: evmInBob and evmSystem* procedures added

# $Id: exec_AAA_Common.tcl 343560 2021-06-16 11:06:03Z smysore $

# getParamValue
#
# returns value of a keyword in a given template, or a
# default value if not found
#
proc getParamValue { template keyword default_value } {

    # Get value of keyword in the template
    set value [evmParameterValue $template $keyword]

    # If value is empty (keyword not found)
    if { [string length $value] == 0 } {

	# Set the result to the specified default value
	set value $default_value

    }

    # Return result
    return $value

}

# hasParamValue
#
# tests if a template contains a given keyword
proc hasParamValue { template keyword } {

    # Get value of keyword in the template
    set value [evmParameterValue $template $keyword]

    # If value is not empty (keyword found)
    if { [string length $value > 0] } {

	# Return "true"
	return "true"

    } else {

	# Return "false"
	return "false"

    }

}

# max:
#
# simple procedure that returns the maximum of N
# numerical values.
#
proc max { args } {

    return [lindex [lsort -real -increasing $args] end]

}

# min:
#
# simple procedure that returns the minimum of two
# values.
#
proc min { args } {

    return [lindex [lsort -real -increasing $args] 0]

}

# FindEnclosingEntries --
#    Search within a sorted list
#
# Arguments:
#    val      Value to be searched
#    values   List of values to be examined
#
# Results:
#    Returns a list of the previous and next indices
#
proc FindEnclosingEntries { val values } {
    set found 0
    set row2  0
    foreach v $values {
	if { $val <= $v } {
	    set row   [expr {$row2-1}]
	    set found 1
	    break
	}
	incr row2
    }

    #
    # Border cases: extrapolation needed
    #
    if { ! $found } {
	incr row2 -1
	set  row $row2
    }
    if { $row == -1 } {
	set row $row2
    }

    return [list $row $row2]
}

# lin_interp
#
# simple linear 1-D interpolation
#
# args: table = name of table containing values
#               values are extracted from the index
#               of the table, considering only the characters
#               before the first comma.
#               Example: index="aaa,bb,1" --> value=aaa
#           x = input value
#
# returns: value based on table, interpolated at x
#
proc lin_interp { table x_val } {

    upvar $table table_l
    set vals [array names table_l]

    set valRange {}

    foreach v $vals {
	regexp {^[^,]+} $v x
	lappend valRange $x
    }

    set valRange [lsort -real $valRange]

    if { [lindex $valRange 0] > $x_val } {
	return $table_l([array names table_l "[lindex $valRange 0],*"])
    }

    if { [lindex $valRange end] < $x_val } {
	return $table_l([array names table_l "[lindex $valRange end],*"])
    }

    foreach {row_min row_max} [FindEnclosingEntries $x_val $valRange] break

    set x_min [lindex $valRange $row_min]
    set x_max [lindex $valRange $row_max]

    set lowIdx [array names table_l "$x_min,*"]
    set highIdx [array names table_l "$x_max,*"]

    set y_min $table_l($lowIdx)
    set y_max $table_l($highIdx)

    if { $x_min != $x_val } {
	set y_val [expr {$y_min+($y_max-$y_min)*($x_val-$x_min)/($x_max-$x_min)}]
    } else {
	set y_val $y_min
    }

    return $y_val
}

# Procedure: wrapList
# arguments:
#  - list: list of elements
#  - length: desired length
#
# returns: list of desired length based on provided list. The original
# list is either truncated or "wrapped around" to produce a list of the
# desired length.
proc wrapList { list length } {

    set l [llength $list]
    
    if { $length <= $l } {

	return [lrange $list 0 [expr $length - 1]]

    } else {

	set nbRepeat [expr $length / $l]
	set list2 $list
	for { set i 1 } { $i <= $nbRepeat } {incr i} {
	    set list2 [concat $list2 $list]
	}

	return [lrange $list2 0 [expr $length - 1]]

    }

}

# 2016-06 (RSC) - BOB support
#
# new procedures to keep track of the system state
# A new global array called 'evmSystemParameters' is used
# to store values of keywords that need to be remembered
# when the ETRM needs to use parameters from different templates
# (e.g. when computing an overhead linked to the change of a filter
# in between two consecutive templates)

# Procedure: evmInBob
# arguments: nothing
# returns  : boolean flag indicating if we
#	     are running in bob or not
#
# Indicates if we are running in bob or not.
#
proc evmInBob {} {
    if { [info proc bobDbRead] == "bobDbRead" } {
	return 1
    } else {
	return 0
    }
}

# Procedure: evmSystemResetParameters
# arguments: list of parameter names
# returns  : nothing
#
# reset and initialize system parameters
#
proc evmSystemResetParameters { {keyList ""} } {
    global evmSystemParameters

    catch { unset evmSystemParameters }
    foreach key $keyList {
	set value "not-found"
	set evmSystemParameters($key) $value
    }
}

# Procedure: evmSystemParameterValue
# arguments: parameter name
# returns  : parameter value
#
# retrieve a system parameter value
#
proc evmSystemParameterValue { key } {
    global evmSystemParameters

    set value "not-found"
    if { [info exists evmSystemParameters($key)] } {
	set value $evmSystemParameters($key)
    }
    return $value
}

# Procedure: evmSystemSetParameterValue
# arguments: parameter name and value
# returns  : nothing
#
# set a system parameter
#
proc evmSystemSetParameterValue { key value } {
    global evmSystemParameters

    set evmSystemParameters($key) $value
}


# Procedure: retrieveAndParse

#
# Retrieves the IDs and Java handles for all containers and OBs
# pertaining to the user's selection in P2PP. The data is stored
# in global variables.
# Each considered OB is also parsed into a global array.
#
# EXAMPLE: with one OB selected within a 3-OB container and one loose OB,
# like depicted below:
#
# +--+ TIME_LINK: id=119 / Name=Matter of time
# |  |
# |  +-- OB: id=121 / Name=the beginning
# |  +-- OB: id=124 / Name=the follow-up  SELECTED
# |  +-- OB: id=127 / Name=the end
# |
# +--+ OB: id=134 / Name=OB at the bottom  SELECTED
#
#
# the compound list will look like this;
#
# {{TIME_LINK 119 java0x27 0} {{OB 121 java0x28 1} {OB 124 java0x29 0} {OB 127 java0x2a 1}}} {OB 134 java0x3b}
#
proc retrieveAndParse {} {

    global scriptingBox compoundList compoundListObjects OBdata

    set parsedOB_ids {}
    set parsedContainer_ids {}
    set selectedOBList {}
    set selectedOB_ids {}
    set selectedContainerList {}
    set selectedContainer_ids {}

    set compoundList {}
    set compoundListObjects {}
    
    # First we get the list of all selected OBs from the API
    # Get total number of selected OBs
    set numOBs [evmGetOBCount]
    
    set obNum 0
    while { $obNum < $numOBs } {
	    
	set ob [evmGetOB $obNum]
	set obID [evmID $ob]
	
	lappend selectedOBList [list $ob $obID]
	lappend selectedOB_ids $obID
	incr obNum
    }
    
    # Same thing with the list of selected containers	
    # Get total number of OB containers present in the selection
    # Note that we deliberately skip 'folders' because they are without
    # any operational meaning.
    set numContainers [evmGetContainerCount]
    
    set containerNum 0
    while { $containerNum < $numContainers } {
	
	set container [evmGetContainer $containerNum]
	
	# Handling of exceptions:
	#  - folders are simply excluded
	#  - empty containers are excluded
	if { [evmType $container] == "FOLDER" || [$container getNumObsBlocks] == 0} {incr containerNum ; continue}

	set containerID [evmID $container]
	lappend selectedContainerList [list $container $containerID]
	lappend selectedContainer_ids $containerID

	incr containerNum
    }

    # Part I. Loop on selected OBs
    
    foreach obEntry $selectedOBList {

	set ob [lindex $obEntry 0]
	set obID [lindex $obEntry 1]
	if { [lsearch -exact $parsedOB_ids $obID] != -1 } { continue }
	
	# Is this OB part of a container?
	set parent [$ob getParentContainer]

	if { ![evmValid $parent] } {

	    # No parent? This is a loose OB. Parse it, store it.
	    # Parse it unless we have done it before
	    if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		lappend parsedOB_ids $obID
		parseOB $ob
	    }

	    lappend compoundList [list OB $obID $ob 1]
	    lappend compoundListObjects [list OB $ob 1]
	    
	} else {

	    # OB is in a container. We parse the entire container.
	    set containerList {}
	    set containerListObjects {}
	    
	    # Collect info about container and store in list
	    set containerID [evmID $parent]
	    set containerType [$parent getType]
	    set containerSelected 0
	    if { [lsearch -exact $selectedContainer_ids $containerID] != -1 } {
		set containerSelected 1
	    }
	    set containerList [list $containerType $containerID $parent $containerSelected]
	    set containerListObjects [list $containerType $parent $containerSelected]
	    
	    # How many OBs does the container have
	    set nbOBs [$parent getNumObsBlocks]

	    set obList {}
	    set obListObjects {}
	    
	    # Loop through all OBs of the container
	    for {set i 0} {$i < $nbOBs} {incr i} {

		# get OB #i (Java pointer)
		set ob [$parent getObsBlock $i]

		# get OB ID
		set obID [evmID $ob]

		set obSelected 0
		if { [lsearch -exact $selectedOB_ids $obID] != -1 } {
		    set obSelected 1
		}
		
		# Parse the OB unless we have done it before
		if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		    lappend parsedOB_ids $obID
		    parseOB $ob
		}

		lappend obList [list OB $obID $ob $obSelected]
		lappend obListObjects [list OB $ob $obSelected]

	    }
	    
	    lappend parsedContainer_ids $containerID
	    
	    lappend compoundList [list $containerList $obList]
	    lappend compoundListObjects [list $containerListObjects $obListObjects]
	    
	}
	
    } ; # end of loop on selected OBs

    # Part II. Loop on selected containers

    foreach containerEntry $selectedContainerList {

	set container [lindex $containerEntry 0]
	set containerID [lindex $containerEntry 1]
	if { [lsearch -exact $parsedContainer_ids $containerID] != -1 } { continue }
	
	set containerType [evmType $container]
	set containerSelected 1
	
	set containerList [list $containerType $containerID $container $containerSelected]
	set containerListObjects [list $containerType $container $containerSelected]
	    
	# How many OBs does the container have
	set nbOBs [$container getNumObsBlocks]

	set obList {}
	set obListObjects {}
	
	# Loop through all OBs of the container
	for {set i 0} {$i < $nbOBs} {incr i} {
	    
	    # get OB #i (Java pointer)
	    set ob [$container getObsBlock $i]
	    
	    # get OB ID
	    set obID [evmID $ob]
	    
	    # NOTE: when selecting a container and computing its execution time, one ALWAYS wants
	    # to consider all OBs inside the container
	    set obSelected 1
	    
	    # Parse the OB unless we have done it before
	    if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		lappend parsedOB_ids $obID
		parseOB $ob
	    }
	    
	    lappend obList [list OB $obID $ob $obSelected]
	    lappend obListObjects [list OB $ob $obSelected]
	    
	}
		
	lappend parsedContainer_ids $containerID
	
	lappend compoundList [list $containerList $obList]
	lappend compoundListObjects [list $containerListObjects $obListObjects]
	
    } ; # end of loop on selected containers
	
    # We store the total number of OBs we parsed
    # (independently of whether they are explicitly selected or not)
    set OBdata(nbOBs) [llength $parsedOB_ids]

    
}


# Procedure: retrieveAndParse300

#
# Retrieves the IDs and Java handles for the OB or the container and its immediate
# children
#
# EXAMPLE: for a 3-OB container like depicted below:
#
# +--+ TIME_LINK: id=119 / Name=Matter of time
#    |
#    +-- OB: id=121 / Name=the beginning
#    +-- OB: id=124 / Name=the follow-up
#    +-- OB: id=127 / Name=the end
#
# the compound list will look like this:
#
# {{TIME_LINK 119 java0x27 1} {{OB 121 java0x28 0} {OB 124 java0x29 0} {OB 127 java0x2a 0}}}
#
proc retrieveAndParse300 {} {

	global OBdata compoundList compoundListObjects

	set compoundList {}
	set compoundListObjects {}

	# Case of OB(s):
	# NOTE: when the EVM is called in p2, it is called for ONE OB at a time.
	# so the loop here is in principle not necessary...
	if { [set nbOBs [evmGetOBCount]] > 0} {

		#puts "DEBUG: retrieveAndParse300 - case of OB"
		set obNum 0
		while {$obNum < $nbOBs } {
			set ob [evmGetOB $obNum]

			if { [evmValid $ob] } {
				set obID [evmID $ob]

				parseOB $ob

				lappend compoundList [list OB $obID $ob 1]
				lappend compoundListObjects [list OB $ob 1]

			}
			incr obNum
		}

		set OBdata(nbOBs) $obNum
	}

	# Case of container(s):
	# NOTE: when the EVM is called in p2, it is called for ONE Container at a time.
	# so the loop here is in principle not necessary...

	set containerTypes { "TIME_LINK" "CONCATENATION" "GROUP" }

	if { [set nbContainers [evmGetContainerCount]] > 0 } {
		#puts "DEBUG: retrieveAndParse300 - case of container #containers----- $nbContainers"

		set containerNum 0
		while { $containerNum < $nbContainers } {

			set container [evmGetContainer $containerNum]

			if { [evmValid $container] && [set nbItems [evmGetItemCount $container]] > 0 && \
				[lsearch -exact $containerTypes [evmType $container]] != -1 } {

				set containerID [evmID $container]
				set containerType [evmType $container]
				#puts "DEBUG: container type $containerID - $containerType"

				lappend compoundList [list $containerType $containerID $container 1]
				lappend compoundListObjects [list $containerType $container 1]

				# loop on elements inside container:
				set itemNum 0
				set itemList {}
				set itemListObjects {}
					
			        #puts "DEBUG: retrieveAndParse300 - case of container #containers----- $nbContainers"	
			        #puts "DEBUG: retrieveAndParse300 - case of container #items----- $nbItems"
			       
				while { $itemNum < $nbItems } {

					set item [evmGetItem $container $itemNum]
					set itemType [evmType $item]
					set itemID [evmID $item]
					
					#puts "DEBUG: retrieveAndParse300 - case of container #item----- $item , $itemType ,  $itemID"

					# IMPORTANT: because verification is done from the bottom-up, the items inside the containers
					# must never be marked as 'selected'

					# Is the item an OB ? Then parse the OB for only the checks relevant for the container-level checks
					if { $itemType == "O" || $itemType == "C" } {

						lappend itemList [list OB $itemID $item 0]
						lappend itemListObjects [list OB $item 0]

						# We parse the OB *only* for parameters needed for the container-level checks
						parseOB $item 1

					} else {
						# Is the item a container ? Yes, then record required characteristics of container... (IS THIS EVER NEEDED??)

						lappend itemList [list $itemType $itemID $item 0]
						lappend itemListObjects [list $itemType $item 0]

					}

					incr itemNum
				}

				lappend compoundList $itemList
				lappend compoundListObjects $itemListObjects

			}

			incr containerNum
		}

		set compoundList [list $compoundList]
		set compoundListObjects [list $compoundListObjects]
	}

	#puts "DEBUG: end retrieveAndParse300 - $compoundList"
}

