# File: exec_OMEGACAM.tcl
# Purpose: computation of execution/exposure time for OMEGACAM OBs
# Author: Stephane Marteau

# $Id: exec_OMEGACAM.tcl 328278 2019-11-29 14:00:38Z smarteau $

#
# Define global parameters for use within following tests
#

#Definition of the fixed overheads
#---------------------------------
#
#   Constant          |  Value (s) | Description
#  -------------------+------------+------------------------------------
#   READOUT_TIME      |  44        | Readout + Data written to disk
#   READOUT_TIME_LAST |  0         | Readout + Data written to disk for last image of an OB
#   FILTER_SETUP      |  65/115    | Filter setup in acquisition and between acquisition and 1st science template
#   FILTER_SETUP_shrt |  35/85     | Filter setup between science templates
#   TEL_PRESET        |  120       | Slew_speed is around 1.5 deg/s. 
#   TEL_OFFSET        |  15        | [probably always absorbed in READOUT_TIME]
#   INS_GUIDESTAR     |  60        | Total time for GS at the OB beginning.
#   INS_IASTAR        |  180       | Time for IA at the OB beginning.
#   INS_GUIDESTAR_NEW |  60        | Acquire new GS after offset 
#   INS_GUIDESTAR_RE  |  5         | Automatically reacquire same guide star after offset.
#   PICK_TIME         |  45        | Time for pick object for acquisition move-to-gap.
#   ADC_SET           |  160       | Time for deploying the ADC
#   
#   LAMP_TIME         |   5        | Time to switch on the lamp
#   EXTRA_SETUP       |   240      | EXTRA_SETUP is 240 seconds for the first exposure in the template if the
#                     |            | previous template executed was *not* one of:
#                     |            | OMEGACAM_img_cal_shutter, OMEGACAM_img_cal_gain, OMEGACAM_img_cal_domeflat,
#                     |            | OMEGACAM_img_cal_quick
 

set execCommon(READOUT)		   44
set execCommon(READOUT_TIME_LAST)   0
set execCommon(TEL_PRESET)	  120
set execCommon(TEL_OFFSET)	   15
set execCommon(INS_GUIDESTAR)	   60
set execCommon(INS_IASTAR)        180
set execCommon(INS_GUIDESTAR_NEW)  60
set execCommon(INS_GUIDESTAR_RE)    5
set execCommon(PICK_TIME)          45
set execCommon(ADC_SET)           160

set execCommon(LAMP_TIME)           5
set execCommon(EXTRA_SETUP)       240

set execCommon(customPresetTime) 0
set execCommon(rotatorOverhead) 0

# OB data - global array containing values of all needed parameters for all templates
# of all OBs.
#
# OBdata(<ob_id>,<tpl_idx>,<keyword>) = <value>
#

global OBdata
array set OBdata {}


# computeDistance
#
# Procedure: computes the angular distance between two points
#            of the sky
#
# inputs:
#   - ra1 : right ascension of the first point, given in arcseconds
#   - ra2 : right ascension of the second point, given in arcseconds
#   - dec1: declination of the first point, given in arcseconds
#   - dec2: declination of the second point, given in arcseconds
#
# output:
#  returns the angular distance in arcseconds
#
proc computeDistance { ra1 ra2 dec1 dec2 } {

    # Conversion factor from arcseconds to radians
    set arcsec2rad [expr 3.14159265358979 / (3600. *180.)]

    # convert input coordinates to radians
    set ra1 [expr $ra1 * $arcsec2rad]
    set ra2 [expr $ra2 * $arcsec2rad]
    set dec1 [expr $dec1 * $arcsec2rad]
    set dec2 [expr $dec2 * $arcsec2rad]

    # Compute angular distance:
    # cos(theta) = sin(dec1)*sin(dec2) + cos(dec1)*cos(dec2)*cos(ra2-ra1)
    #

    set t [expr sin($dec1)*sin($dec2) + cos($dec1)*cos($dec2)*cos($ra2-$ra1)]
    set dist [expr acos($t)]

    # Convert to arcseconds
    set dist [expr abs($dist / $arcsec2rad)]

    return $dist
}

# getPresetTime
#
# Procedure: computes the time needed to go from one telescope
# preset to another, taking into account the distance between
# the two pointings.
#
# inputs:
#   - ob1, ob2: OB objects
#
# output:
#  returns the time overhead in seconds
#
proc getPresetTime { ob1 ob2 } {

    global execCommon

    set type1 [evmType $ob1]
    set type2 [evmType $ob2]

    if { $type1 == "C" && $type2 == "O" } {
	return $execCommon(TEL_PRESET)
    } elseif { $type1 == "O" && $type2 == "C" } {
	return 0
    }
    # Targets from both OBs
    set target1 [evmTarget $ob1]
    set target2 [evmTarget $ob2]

    # Get right ascension + declination from target packages
    set ra1 [expr [evmRightAscension $target1 marcsec] / 1000.]
    set dec1 [expr [evmDeclination $target1 marcsec] / 1000.]
    set ra2 [expr [evmRightAscension $target2 marcsec] / 1000.]
    set dec2 [expr [evmDeclination $target2 marcsec] / 1000.]

    # Compute angular distance between targets
    set distance [computeDistance $ra1 $ra2 $dec1 $dec2]

    set slewSpeed 5400 ; # slew speed = 1.5 deg/s = 5400 arcsec/s

    # Overhead = 30 seconds + slew time
    set preset_time [expr 30 + $distance/$slewSpeed]
    return $preset_time
}


proc getTelescopePreset { ob } {

    global OBdata execCommon

    set ob_id [evmID $ob]
    set theta $OBdata($ob_id,0,TEL.ROT.OFFANGLE)
    if { $theta > 180 } { set theta [expr $theta - 360] }
    if { $theta < -180 } { set theta [expr $theta + 360] }
    
    set rot_speed 2.0 ; # average rotator speed in deg/s
    return [expr $execCommon(TEL_PRESET) + abs($theta)/$rot_speed]

}

proc getRotatorOverhead { ob1 ob2 } {

    global OBdata
	
    set ob_id1 [evmID $ob1]
    set ob_id2 [evmID $ob2]
	
    set angle1 $OBdata($ob_id1,0,TEL.ROT.OFFANGLE)
    set angle2 $OBdata($ob_id2,0,TEL.ROT.OFFANGLE)
	
    foreach theta [list angle1 angle2] {
	    
	if { [set $theta] > 180 } { set $theta [expr [set $theta] - 360] }
	if { [set $theta] < -180 } { set $theta [expr [set $theta] + 360] }
    }

    set rot_speed 2.0 ; # average rotator speed in deg/s
    set rotator_overhead [expr ( abs($angle1) + abs($angle2) ) / $rot_speed ]
    return $rotator_overhead
	
}

# parseOB
#
# Procedure: parses an OB and stores its contents in the global
# table 'OBdata'.
#
# inputs:
#   - ob: OB object (Java handle)
#
# output:
#   contents of OB are stored in global table 'OBdata'
#
proc parseOB { ob } {

    global OBdata scriptingBox

    set ob_id [evmID $ob]

    # List of all templates (Java handles)
    set templates [evmTemplates $ob]

    # template counter (acquisition=0)
    set index 0

    # Loop on all templates
    foreach temp $templates {

	# Parse all valid templates (verification required to
	# catch exception for 1st template of CalBlocks which is always NULL)
	if { [evmValid $temp] } {
	
	    # Record template name and type
	    #set OBdata($ob_id,$index,tplType) [evmTemplateType $temp]
	    set OBdata($ob_id,$index,tplName) [evmName $temp]

	    # Parse individual template
	    parseTemplate $temp $ob_id $index

	    incr index ; # template counter: add 1
	}

    }

    # Store how many templates this OB has
    set OBdata($ob_id,nbTpl) $index

    return
}


# Procedure: parseTemplate
# arguments:
#  - template: EVM template object
#
# returns: nothing (global array is used)
#
# Parses a template and puts the parameters into
# a global array
#
proc parseTemplate { template ob_id idx } {

    # Global array in which parameters are stored
    global OBdata

    # List of keywords to retrieve + default values to assign in the code
    # when they do not exist in the given template.
    set keywordList { "DET1.WIN1.UIT1" "-1" "SEQ.NEXPO" "-1" "INS.FILT.NAME" "not-found" \
			  "TEL.TARG.ALPHA" "" "TEL.TARG.DELTA" "" "TEL.ROT.OFFANGLE" "0" \
	                  "TEL.TARG.OFFSETSIZE" "0" \
	                  "TEL.TARG.OFFSETSIZEX" "0" "TEL.TARG.OFFSETSIZEY" "0" \
			  "TEL.TARG.DX" "0" "TEL.TARG.DY" "0" \
			  "TEL.TARG.OFFSETALPHA" {} "TEL.TARG.OFFSETDELTA" {} \
			  "OCS.AG.START" "" "OCS.IA.START" "" "TEL.ADC.TYPE" "" "TEL.TARG.PATTERN" "" \
    }

    # Loop on all keywords to retrieve
    foreach {key def} $keywordList {

	# Retrieve value of keyword from template, assign default value if not found
	set OBdata($ob_id,$idx,$key) [getParamValue $template $key $def]

    }

    return
}


# procedure: getAcquisitionTime
# arguments:
#  - ob_id: ID of the OB containing the template
#  - idx: index of the template within the OB
#
# returns: execution time for acquisition templates of OMEGACAM
#
proc getAcquisitionTime { ob_id idx } {

    # Global variables
    global execCommon OBdata
    global previousOBId

    # We check the value of the autoguiding flag in the acquisition
    # in order to correctly set the value for some fixed overheads
    if { $OBdata($ob_id,$idx,OCS.AG.START) == "F" } {
	    
	set execCommon(INS_GUIDESTAR) 0
	set execCommon(INS_GUIDESTAR_NEW) 0
	set execCommon(INS_GUIDESTAR_RE) 0
	    
    } else {
	    
	set execCommon(INS_GUIDESTAR)	   60
	set execCommon(INS_GUIDESTAR_NEW)  60
	set execCommon(INS_GUIDESTAR_RE)    5
	    
    }

    # If the ADC is not used then set the corresponding overhead to zero
    if { $OBdata($ob_id,$idx,TEL.ADC.TYPE) == "NONE" } {
	    
	set execCommon(ADC_SET) 0
	    
    } else {

	set execCommon(ADC_SET) 160
	    
    }
	
    # If the Image Analysis is not used then set the corresponding overhead to zero
    if { $OBdata($ob_id,$idx,OCS.IA.START) == "F" } {
	    
	set execCommon(INS_IASTAR)  0

    } else {

	set execCommon(INS_IASTAR)  180
	    
    }

    # Preset time: if we compute the slew time then it has to be compared with the
    # time required to switch filters between OBs and to reposition the rotator. Largest overhead wins.
    if { [info exists previousOBId] } {
	    
	set preset [expr [max $execCommon(customPresetTime) $execCommon(FILTER_CHANGE_BETWEEN_OBs)] + $execCommon(rotatorOverhead)]
	
	# In case of concatenated OBs, we do not count any overhead for the ADC in case the
	# previous OB has the ADC set to true.
	if { [info exists OBdata($previousOBId,0,TEL.ADC.TYPE)] && \
		 $OBdata($previousOBId,0,TEL.ADC.TYPE) == "COORD" } {
			
	    set execCommon(ADC_SET) 0
	}
	    
    } else {
	    
	set preset $execCommon(customPresetTime)
	    
    }

    # Store name of acquisition template into global array
    set templateName $OBdata($ob_id,$idx,tplName)
    set execCommon(acqName) $templateName
 
    # Overhead common to all acq. templates
    set executionTime [expr [max $preset $execCommon(ADC_SET)] + $execCommon(INS_GUIDESTAR) + $execCommon(INS_IASTAR)]
	
    # Add specific overhead for 'movetogap' acquisition
    if { $templateName == "OMEGACAM_img_acq_movetogap" } {
	set executionTime [expr $executionTime \
	 	    	     + 1 * ( $OBdata($ob_id,$idx,DET1.WIN1.UIT1) + $execCommon(READOUT) + $execCommon(PICK_TIME) )]
    }
	
    # Return the execution time for the acquisition
    return $executionTime

}


# Procedure:getFilterChangeTime
#
# Returns the time needed to change from one filter to another one.
# This time depends on the relative position of the 2 filters in the 
# two magazines.
#
# Arguments:
#  - filter1: name of current filter
#  - filter2: name of filter to switch to...
#  - isShort: flag to determine if the full length overhead FILTER_SETUP of 65/115 s is 
#    to be used (between aquisition and science templates) or 
#    FILTER_SETUP_shrt of 35/85 s is to be used (between science templates)
#
# Returns:
#  time to switch from filter1 to filter2 (in seconds)
#
proc getFilterChangeTime { filter1 filter2 isShort } {

    # List of filters in both magazines
    set filterMag1 {u_SDSS r_SDSS z_SDSS V_JOHN u_g_r_i_SDSS v_STRM }
    set filterMag2 {opaque g_SDSS i_SDSS B_JOHN NB_659 H_ALPHA }

    if { [lsearch -exact $filterMag1 $filter1] != -1 } {
	set mag1 1
    } elseif { [lsearch -exact $filterMag2 $filter1] != -1 } {
	set mag1 2
    } else {
	set mag1 0
    }

    if { [lsearch -exact $filterMag1 $filter2] != -1 } {
	set mag2 1
    } elseif { [lsearch -exact $filterMag2 $filter2] != -1 } {
	set mag2 2
    } else {
	set mag2 0
    }


    if { $filter1 == $filter2 } {
	set filter_change 0
    } elseif { $mag1 == 0 || $mag2 == 0 } {
	set filter_change 0
    } elseif { $mag1 == $mag2 } {
	if { $isShort == "false" } {
	    set filter_change 115
	} else {
	    set filter_change 85
	}
    } elseif { $mag1 != $mag2 } {
	if { $isShort == "false" } {
	    set filter_change 65
	} else {
	    set filter_change 35
	}	
    }

    # Get the overhead from the above table
    return $filter_change

}


# Procedure: getFilterChangeOverhead
# arguments:
#  - ob_id : OB ID
#  - tpl_idx: index of template 
#
# returns: overhead due to change of filters (in seconds)
#
# There is an overhead in between two templates whenever the
# last filter of template 1 is not the same as the first
# filter of template 2.
# Takes into account the relative positions of filters to
# compute the overhead.
# If one of the two filters passed as argument is not in the list
# of known filters, the returned value is zero.
#
proc getFilterChangeOverhead { ob_id tpl_idx } {

    global execCommon OBdata

    set overhead 0
    set isShort "true"

    # when 1st template is aquisition and 2nd template is science
    set lastTemplateName [evmSystemParameterValue "lastTemplateName"]
    if { [regexp {acq} $lastTemplateName] } {
	set isShort "false"
    }
    
    # List of filters used in templates 1 and 2
    set filterTemplate $OBdata($ob_id,$tpl_idx,INS.FILT.NAME)
	
    set lastFilter [evmSystemParameterValue "INS.FILT.NAME"]

    # Check that the 2 templates DO have filters...
    # ... then get the filters of template1 and template2
    if {$filterTemplate != "not-found" && \
	    $lastFilter != "not-found" && \
	    $filterTemplate != $lastFilter } {
	
	# Get the associated overhead
	set overhead [getFilterChangeTime $lastFilter $filterTemplate $isShort]
	
    }

    return $overhead

}

# computeFilterChangeBetweenOBs
#
# Procedure: compute the overhead linked to switching filters in between
# OBs.
# 
# inputs:
#   - id1, id2: IDs of concerned OBs
#
# output:
#   sets a global variable with the overhead between the 2 OBs in seconds
#
proc computeFilterChangeBetweenOBs { id1 id2 } {

    global execCommon OBdata

    # Find last filter in OB1: start from last template and go back until
    # a filter is found.
    set filter1 ""
    for { set i [expr $OBdata($id1,nbTpl) - 1] } { $i >= 0 } { incr i -1} {

	if { $OBdata($id1,$i,INS.FILT.NAME) != "" } {
	    set filter1 $OBdata($id1,$i,INS.FILT.NAME)
	    break
	}

    }


    # If no filter found in OB #1, overhead=0 --> exit
    if { $filter1 == "" } { set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0 ; return }

    # Find first filter in OB2: start from first template and go
    # forward until a filter is found.
    set filter2 ""
    for { set i 0 } { $i < $OBdata($id2,nbTpl) } { incr i } {

	if { $OBdata($id2,$i,INS.FILT.NAME) != "" } {
	    set filter2 $OBdata($id2,$i,INS.FILT.NAME)
	    break
	}

    }


    # If no filter found in OB #2, overhead=0 --> exit
    if { $filter2 == "" } { set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0 ; return }

    #between OBs always use the full length overhead FILTER_SETUP of 65/115 s
    set isShort "false"
    # Compute the overhead based on the filters that were found
    return [getFilterChangeTime $filter1 $filter2 $isShort]

}

# Procedure: AbsToRelOffsets
#
# Takes a X+Y list of list of absolute offsets and returns
# the same but as relative offsets
proc AbsToRelOffsets { offsLists } {

    # Transform a list of absolute offsets into relative ones
    proc doAbsToRelOffsets { absOffs } {
        set relOffs [lindex $absOffs 0]
        for { set i 1 } { $i < [llength $absOffs] } { incr i } {
            set cur [lindex $absOffs $i]
            set prev [lindex $absOffs [expr {$i-1}]]
        lappend relOffs [expr {$cur-$prev}]
        }
        return $relOffs
    }

    set listRA  [doAbsToRelOffsets [lindex $offsLists 0]]
    set listDEC [doAbsToRelOffsets [lindex $offsLists 1]]
    return [list $listRA $listDEC]
}


# Procedure: computeJitterOffsets
#
# The procedure computes the suite of *relative* offsets used by the
# template img_obs_jitter based on the number of offsets and their
# size. The code is taken directly from the one used for the
# execution of the template itself (courtesy of Andrea Barrufolo)
#
# arguments:
#  - N: number of offsets
#  - offsetSize: size of the individual offsets
#
# returns: list of relative X and Y offsets
#
proc computeJitterOffsets { N offsetSize } {
	
    #--------------------------------------------------------------------
    # Optimize
    #   This function reorders the sequence of offset steps in order to 
    #   improve the overall telescope offsetting time. For each step, it
    #   finds the closest one in the list and takes it as the following 
    #   one. It is not the best approach (which would be some "travelling 
    #   salesman" algorithm, too complicated and lengthy) but it is 
    #   sufficient for our purposes. 
    #
    proc Optimize {lists} {

        set A [lindex $lists 0]
        set B [lindex $lists 1]

        if { [llength $A] != [llength $B]}  {
            error "A and B lists have different lengths"
        }

        set x_LAST 0
        set y_LAST 0

        set X {}
        set Y {}

        while { [llength $A] } {
            set bestDistance 99999999

            for {set i 0} {$i < [llength $A]} {incr i} {
                set xOffs    [expr {[lindex $A $i] - $x_LAST}]
                set yOffs    [expr {[lindex $B $i] - $y_LAST}]
                set distance [expr sqrt(pow($xOffs,2)+pow($yOffs,2))]

                if { $distance < $bestDistance } {
                    set bestIndex    $i
                    set bestDistance $distance
                }
            }

            set x_LAST [lindex $A $bestIndex]
            set y_LAST [lindex $B $bestIndex]
            set A      [lreplace $A $bestIndex $bestIndex]
            set B      [lreplace $B $bestIndex $bestIndex]

            lappend X $x_LAST
            lappend Y $y_LAST
        }

        return [list $X $Y]
    }


    #--------------------------------------------------------------------
    # getOffsJitterList
    #
    proc getOffsJitterList { N offsetSize } {

        set offsALPHA {}
        set offsDELTA {}

	set N4 [expr {int($N/4)}]
	set L [expr {$N/2.0-0.5}]
	set rest [expr {$N % 4}]

	switch $rest {
	    0 { set startI 0 }
	    1 { 
		set OffsRA  0
		set OffsDEC 0
		set startI  1
	    }
	    2 {
		set OffsRA  {0.5 -0.5}
		set OffsDEC $OffsRA
		set startI 2
	    }
	    3 {
		set OffsRA  {1 0 -1}
		set OffsDEC {1 -1 0}
		set startI 3
	    }
	}

	for {set i $startI} {$i<$N} {incr i} {
	    lappend OffsRA  0
	    lappend OffsDEC 0
	}

	set D [expr {0.5*(1+$rest)}]

	for {set i 0} {$i < $N4} {incr i} {
	    if {[expr {$i % 2}] == 0} {
		set a [expr {$L-0.5*$i}]
		set b [expr {$D+1.5*$i}]
	    } else {
		set a [expr {$D+0.5+1.5*$i}]
		set b [expr {$a-1}]
	    }
	    if { [expr {int($a)%2}] == 0 } {set a [expr {-$a}]}
	    set I0 [expr {$startI+$i*4}]
	    set I1 [expr {$I0+3}]
	    set OffsRA  [lreplace $OffsRA  $I0 $I1 $a  $b [expr {-$a}] [expr {-$b}]]
	    set OffsDEC [lreplace $OffsDEC $I0 $I1 $b [expr {-$a}] [expr {-$b}]  $a]
	}

        # Multiply each alpha and delta element by the offset
        for {set i 0} {$i < $N} {incr i} {
            lappend offsALPHA [expr {[lindex $OffsRA $i] * $offsetSize}]
            lappend offsDELTA [expr {[lindex $OffsDEC $i] * $offsetSize}]
        }

	return [list $offsALPHA $offsDELTA]
	   
    }

    set pat [getOffsJitterList $N $offsetSize]
    set pat [Optimize $pat]
    set pat [AbsToRelOffsets $pat]

    return $pat

}

# Procedure: computeDitherDiagOffsets
#
# The procedure computes the suite of *relative* offsets used by the
# template img_obs_dither with the 'diag' pattern, based on the number
# of offsets and their size. The code is taken directly from the one
# used for the execution of the template itself (courtesy of Andrea Barrufolo)
#
# arguments:
#  - N: number of offsets
#  - offsetSizeX: size of the individual offsets along X
#  - offsetSizeY: size of the individual offsets along Y
#
# returns: list of relative X and Y offsets
#
proc computeDitherDiagOffsets { N offsetSizeX offsetSizeY } {

    # Computes the list of absolute offsets for 'diag' dither pattern
    proc getOffsDitherDiagList { N offsetSizeX offsetSizeY } {

      set N_Is_Even [expr {($N%2) ? 0 : 1}]

      if { $N_Is_Even } {

          set OffsALPHA -0.5
          set OffsDELTA -0.5

      } else {

          set OffsALPHA 0.0
          set OffsDELTA 0.0

      }

      set J 1

      set lastALPHA $OffsALPHA
      set lastDELTA $OffsDELTA

      for { set i 1 } { $i < $N } { incr i } {

          set sign [expr {($i%2) ? 1 : -1}]

          set newALPHA [expr {$lastALPHA+$sign*$J*$offsetSizeX}]
          set newDELTA [expr {$lastDELTA+$sign*$J*$offsetSizeY}]

          lappend OffsALPHA $newALPHA
          lappend OffsDELTA $newDELTA

          set lastALPHA $newALPHA
          set lastDELTA $newDELTA

          incr J

      }

      return [list $OffsALPHA $OffsDELTA]
	    
    }

    set xyList [getOffsDitherDiagList $N $offsetSizeX $offsetSizeY]
    set xyList [AbsToRelOffsets $xyList]	

    return $xyList
	
}

# Procedure: computeDitherStarextOffsets
#
# The procedure computes the suite of *relative* offsets used by the
# template img_obs_dither with the 'starext' pattern, based on the number
# of offsets and their size. The code is taken directly from the one
# used for the execution of the template itself (courtesy of Andrea Barrufolo)
#
# arguments:
#  - N: number of offsets
#  - DX: X-size of area to keep unexposed
#  - DY: Y-size of area to keep unexposed
#  - XGap: size of the gap in X to bridge
#  - YGap: size of the gap in Y to bridge
#
# returns: list of relative X and Y offsets
#
proc computeDitherStarextOffsets { N DX DY XGap YGap} {

    # Computes the list of absolute offsets for 'diag' dither pattern
    proc getOffsDitherStarextList { NumOffs DX DY XGap YGap } {

        set baseOffsRA 0
	
        set tmp [expr {$XGap*1.5-$DX*0.5}]

        lappend baseOffsRA  $tmp
        lappend baseOffsRA [expr {-$tmp}]

        set tmp [expr {$XGap*0.5-$DX*0.5}]

        lappend baseOffsRA  $tmp
        lappend baseOffsRA [expr {-$tmp}]

        set baseOffsDEC 0

        set tmp [expr {$YGap*0.5-$DY*0.5}]

        lappend baseOffsDEC  $tmp
        lappend baseOffsDEC [expr {-$tmp}]

        set tmp [expr {-$YGap*1.5+$DY*0.5}]

        lappend baseOffsDEC  $tmp
        lappend baseOffsDEC [expr {-$tmp}]

        set corr [expr {0.5*int(($NumOffs-1)/5)}]

        for {set i 0} {$i<$NumOffs} {incr i} {

            set I0 [expr {$i % 5}]
	      
            set Add [expr {int($i/5)-$corr}]

            set tmp [lindex $baseOffsRA  $I0]
            lappend OffsRA  [expr {$tmp+$Add}]

            set tmp [lindex $baseOffsDEC $I0]
            lappend OffsDEC [expr {$tmp+$Add}]

        }

        return [list $OffsRA $OffsDEC]

    }

    set xyList [getOffsDitherStarextList $N $DX $DY $XGap $YGap]
    set xyList [AbsToRelOffsets $xyList]	

    return $xyList
	
}



# Procedure: getExpoExecTime
#
# Top level procedure for exposure/execution time, called from SetExec.
# Computes exposure and execution time for a given science template.
# arguments:
#  - ob_id: ID of the OB that contains the template
#  - tpl_idx: index of the template within the OB
#
# returns: list of exposure + execution time (in seconds)
#
proc getExpoExecTime { ob_id tpl_idx } {

    global execCommon OBdata

    # Reset times for the template
    set exposureTime 0
    set executionTime 0

    # Get template name
    set templateName $OBdata($ob_id,$tpl_idx,tplName)

    # List of science templates.
    set tplList { "OMEGACAM_img_obs_dither" "OMEGACAM_img_obs_stare" \
	    "OMEGACAM_img_obs_jitter" "OMEGACAM_img_obs_offset" "OMEGACAM_img_cal_zp" "OMEGACAM_img_cal_monit" }

    if { [lsearch -exact $tplList $templateName] != - 1 } {

	# Maximum dither pattern to check offsets against
	set maxDither_x 50
	set maxDither_y 170

	if { $templateName == "OMEGACAM_img_obs_dither" } {

	    set exposureTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]

	    # Compute list of relative offsets
	    switch $OBdata($ob_id,$tpl_idx,TEL.TARG.PATTERN) {

		"diag" {
		    set offsets [computeDitherDiagOffsets \
			$OBdata($ob_id,$tpl_idx,SEQ.NEXPO) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZEX) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZEY) ]
		}

		"starext" {
		    set offsets [computeDitherStarextOffsets \
			$OBdata($ob_id,$tpl_idx,SEQ.NEXPO) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.DX) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.DY) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZEX) \
			$OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZEY) ]
		}
	    }

	    set offsetList_x [lindex $offsets 0]
	    set offsetList_y [lindex $offsets 1]

	    set offset_c_x 0
	    set offset_c_y 0

	    for { set i 0 } { $i < $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) } { incr i } {

		set offset_x [lindex $offsetList_x $i]
		set offset_y [lindex $offsetList_y $i]

		set offset_c_x [expr $offset_c_x + $offset_x]     
		set offset_c_y [expr $offset_c_y + $offset_y]     


		if { [expr abs($offset_c_x)] > $maxDither_x || [expr abs($offset_c_y)] > $maxDither_y } {

		    # Re-acquire guide star: count additional overhead
		    set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_NEW)

		    # Reset cumulative offsets
		    set offset_c_x 0
		    set offset_c_y 0

		} else {

		    set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_RE)
		}

		set executionTime [expr $executionTime + \
		    $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) + $INS_GUIDESTAR_NEW ]
	    }


	    # NB: the overhead linked to changing filters between templates is accounted for
	    # in the procedure 'setExecTimeForOneOB'.

	}

	if { $templateName == "OMEGACAM_img_obs_stare" } {

	    set exposureTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]


	    set executionTime [expr ( $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) ) *  $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) \
		+ ( $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) - 1) * $execCommon(INS_GUIDESTAR_RE) ]

	    # NB: the overhead linked to changing filters between templates is accounted for
	    # in the procedure 'setExecTimeForOneOB'.

	}


	if { $templateName == "OMEGACAM_img_obs_jitter" } {

	    set exposureTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]

	    if { [expr ($OBdata($ob_id,$tpl_idx,SEQ.NEXPO) -1) * $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZE)] < 100 } {

		set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_RE)

		set executionTime [expr $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * ($OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) + $INS_GUIDESTAR_NEW)]

	    } else {

		# Compute list of relative offsets
		set offsets [ computeJitterOffsets \
		    $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) \
		    $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETSIZE) ]

		set offsetList_x [lindex $offsets 0]
		set offsetList_y [lindex $offsets 1]

		set offset_c_x 0
		set offset_c_y 0

		for { set i 0 } { $i < $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) } { incr i } {

		    set offset_x [lindex $offsetList_x $i]
		    set offset_y [lindex $offsetList_y $i]

		    set offset_c_x [expr $offset_c_x + $offset_x]     
		    set offset_c_y [expr $offset_c_y + $offset_y]     


		    if { [expr abs($offset_c_x)] > $maxDither_x || [expr abs($offset_c_y)] > $maxDither_y } {

			# Re-acquire guide star: count additional overhead
			set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_NEW)

			# Reset cumulative offsets
			set offset_c_x 0
			set offset_c_y 0

		    } else {

			set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_RE)
		    }

		    set executionTime [expr $executionTime + \
			$OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) + $INS_GUIDESTAR_NEW ]
		}
	    }

	    # NB: the overhead linked to changing filters between templates is accounted for
	    # in the procedure 'setExecTimeForOneOB'.

	}


	if { $templateName == "OMEGACAM_img_obs_offset" } {

	    set exposureTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]

	    # The loop below extends the offset lists if necessary.
	    if { [set nbMissing [expr $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) - [llength $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA)]]] > 0 } {

		set nbOffset [llength $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA)]
		set idx 0

		for {set i 1} {$i <= $nbMissing} {incr i} {

		    lappend OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA) [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA) $idx]

		    incr idx

		    if {$idx > [expr $nbOffset - 1] } { set idx 0 }

		}
	    }

	    if { [set nbMissing [expr $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) - [llength $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA)]]] > 0 } {

		set nbOffset [llength $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA)]
		set idx 0

		for {set i 1} {$i <= $nbMissing} {incr i} {

		    lappend OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA) [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA) $idx]

		    incr idx

		    if {$idx > [expr $nbOffset - 1] } { set idx 0 }

		}
	    }

	    set offset_c_x 0
	    set offset_c_y 0
	    set rotatorAngle [expr $OBdata($ob_id,0,TEL.ROT.OFFANGLE) * 3.14159265358979 / 180.]

	    for { set i 1 } { $i <= $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) } { incr i } {

		set offsetAlpha [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA) [expr $i-1] ]
		set offsetDelta [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA) [expr $i-1] ]

		set offset_c_x [expr $offset_c_x + \
		    cos($rotatorAngle) * $offsetAlpha - sin($rotatorAngle) * $offsetDelta ]

		set offset_c_y [expr $offset_c_y + \
		    sin($rotatorAngle) * $offsetAlpha + cos($rotatorAngle) * $offsetDelta ]

		if { [expr abs($offset_c_x)] > $maxDither_x || [expr abs($offset_c_y)] > $maxDither_y } {

		    set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_NEW)       
		    set offset_c_x 0
		    set offset_c_y 0

		} else {

		    set INS_GUIDESTAR_NEW $execCommon(INS_GUIDESTAR_RE)
		}

		set executionTime [expr $executionTime + \
		    $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) + $INS_GUIDESTAR_NEW
		]
	    }

	    # NB: the overhead linked to changing filters between templates is accounted for
	    # in the procedure 'setExecTimeForOneOB'.

	}

	if { $templateName == "OMEGACAM_img_cal_zp" || $templateName == "OMEGACAM_img_cal_monit" } {

	    set exposureTime $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1)
	    set executionTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT)]

	    # NB: the overhead linked to changing filters between templates is accounted for
	    # in the procedure 'setExecTimeForOneOB'.

	}
	if { $templateName == "OMEGACAM_img_obs_offset" } {
	    #If in OMEGACAM_img_obs_offset, the first RA offset is !=0 || first Dec offset !=0, add TEL_OFFSET to the execution time of that template.
	    set firstRAoffset [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETALPHA) 0]
	    set firstDECoffset [lindex $OBdata($ob_id,$tpl_idx,TEL.TARG.OFFSETDELTA) 0]
	    #puts "firstRAoffset $firstRAoffset firstDECoffset $firstDECoffset"
	    if { $firstRAoffset != 0 || $firstDECoffset !=0 } {
		set executionTime [expr $executionTime + $execCommon(TEL_OFFSET)]
	    }		
	}

    }


    ###
    # DAYTIME CALIBRATIONS
    ###

    set lastTemplateName [evmSystemParameterValue "lastTemplateName"]
	
    if { $templateName == "OMEGACAM_img_cal_bias" } {

	set exposureTime 0
	set executionTime [expr $execCommon(READOUT) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]

    }

    if { $templateName == "OMEGACAM_img_cal_dark" } {

	set exposureTime [expr $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) ]
	set executionTime [expr ( $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) + $execCommon(READOUT) ) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) ]

    }
	
    if { $templateName == "OMEGACAM_img_cal_domeflat" } {

	set exposureTime 0
	set executionTime 0

	foreach dit $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) {
	    set exposureTime [expr $exposureTime + $dit * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) ]    

	    set executionTime [expr $executionTime + \
		( $execCommon(LAMP_TIME) + $execCommon(READOUT) + $dit ) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)  ]
	}

	set tplList { OMEGACAM_img_cal_shutter OMEGACAM_img_cal_gain OMEGACAM_img_cal_domeflat OMEGACAM_img_cal_quick }

	if { [lsearch -exact $tplList $lastTemplateName] == -1 } {
	    set executionTime [expr $executionTime + $execCommon(EXTRA_SETUP)]
	}

    }
	
    if { $templateName == "OMEGACAM_img_cal_quick" } {

	set exposureTime $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1)	    
	set executionTime [expr $execCommon(LAMP_TIME) + $execCommon(READOUT) + $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) ]

	set tplList { OMEGACAM_img_cal_shutter OMEGACAM_img_cal_gain OMEGACAM_img_cal_domeflat OMEGACAM_img_cal_quick }

	if { [lsearch -exact $tplList $lastTemplateName] == -1 } {
	    set executionTime [expr $executionTime + $execCommon(EXTRA_SETUP)]
	}

    }
	
    if { $templateName == "OMEGACAM_img_cal_readnoise" } {

	set exposureTime 0
	set executionTime [expr $execCommon(READOUT) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) ]

    }
	
    if { $templateName == "OMEGACAM_img_cal_shutter" || $templateName == "OMEGACAM_img_cal_gain" } {

	set expoList [wrapList $OBdata($ob_id,$tpl_idx,DET1.WIN1.UIT1) $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)]
	set exposureTime 0
	foreach dit $expoList {
	    set exposureTime [expr $exposureTime + $dit]
	}

	set executionTime $execCommon(LAMP_TIME)

	for { set i 0 } {$i < $OBdata($ob_id,$tpl_idx,SEQ.NEXPO)} {incr i} {

	    set dit [lindex $expoList $i]

	    set executionTime [expr $executionTime + $execCommon(READOUT) + $dit ]
	}

	set tplList { OMEGACAM_img_cal_shutter OMEGACAM_img_cal_gain OMEGACAM_img_cal_domeflat OMEGACAM_img_cal_quick }

	if { [lsearch -exact $tplList $lastTemplateName] == -1 } {
	    set executionTime [expr $executionTime + $execCommon(EXTRA_SETUP)]
	}

    }
    
	
    # Return back exposure + execution times as a list
    return [list $exposureTime $executionTime]

}

# setExecTimeForOneOB
#
# Procedure: computes the exposure and execution times
# for a single OB
#
# inputs:
#   - objRef: Java handle to OB object
#
# output:
#   none. Uses the EVM API to internally set the exposure and
#   execution time to the OB inside P2PP
#
proc setExecTimeForOneOB { objRef } {

    global OBdata scriptingBox execCommon

    # Get OB ID
    set ob_id [evmID $objRef]

    # Get list of templates
    set templates [evmTemplates $objRef]

    set previousTemplate "null"

    # Init variables
    set exposureTime 0
    set executionTime 0

    # Template index
    set tplIdx 0

    #This will contain the name of the last template of the OB outside the loop on templates
    set templateName ""
    # Loop on all templates
    foreach temp $templates {

	# NB: test validity of template (avoids working on a NULL object
	# with calblocks that do not have an acq template)
	if { [evmValid $temp] } {

	    #NB 23Sep2009: the function evmType does not seem to work
	    # properly, resorting to template name to determine acq vs. science
	    set templateName $OBdata($ob_id,$tplIdx,tplName)

	    if { [regexp {acq} $templateName] } {

		# Acquisition template: get execution time
		set executionTime [expr $executionTime + [getAcquisitionTime $ob_id $tplIdx]]
		set previousTemplate $temp

	    } else {

		if { $previousTemplate != "null" } {

		    # Filter change time (in case of filter change)
		    set filterChangeTime [getFilterChangeOverhead $ob_id $tplIdx]
		    set executionTime [expr $executionTime + $filterChangeTime]
		}

		# Other templates: get exposure + execution times
		set expo_exec_times [getExpoExecTime $ob_id $tplIdx]
		set exposureTime [expr $exposureTime + [lindex $expo_exec_times 0]]
		set executionTime [expr $executionTime + [lindex $expo_exec_times 1]]

		set previousTemplate $temp

	    }

	    # We store the value of the last filter used (if any) for later retrieval
	    if { $OBdata($ob_id,$tplIdx,INS.FILT.NAME) != "not-found" } {
		evmSystemSetParameterValue "INS.FILT.NAME" $OBdata($ob_id,$tplIdx,INS.FILT.NAME)
	    }
		
	    evmSystemSetParameterValue "lastTemplateName" $templateName
    
	    incr tplIdx
	}

    }; # end of loop on templates

    # List of science templates for which the READOUT_LAST = 0.
    set tplList { "OMEGACAM_img_obs_dither" "OMEGACAM_img_obs_stare" \
		"OMEGACAM_img_obs_jitter" "OMEGACAM_img_obs_offset" "OMEGACAM_img_cal_zp" "OMEGACAM_img_cal_monit" }

    if { [lsearch -exact $tplList $templateName] != - 1 } {
	#for the last image the $execCommon(READOUT) === $execCommon(READOUT_LAST) = 0, hence subtract 44 s
	set executionTime [expr $executionTime - $execCommon(READOUT)]
    }

    # Set execution and exposure times in P2PP
    evmSetExecutionTime $objRef [expr round($executionTime)]
    evmSetExposureTime $objRef [expr round($exposureTime)]

    return
}

# Procedure: setExecTime
# arguments: none
# returns  : nothing
#
# Sets the exposure and execution time in P2PP for the selected OB(s)
#
proc setExecTime {} {

    # Global variables
    global scriptingBox
    global execCommon OBdata
    global compoundList compoundListObjects
    global previousOBId
	
    # Initialization of EVM process
    set version [evmInit]

    # initialize the system parameters
    #
    # BOB needs to call this procedure separately,
    # therefore skip it, if we are running in BOB
    if { ![evmInBob] } {
	setSystemParameters
    }

    # Analyze selection of OBs + containers
    # and parse all relevant OBs
    retrieveAndParse

    # Loop over all elements in the compound list
    foreach object $compoundListObjects {
	
	# Get 'object type' (first element of object list)
	set objType [lindex $object 0]
	
	# If the object type is a list, then we are dealing
	# with a container: the actual type has to be fetched
	# deeper inside the object list
	if { [llength $objType] > 1 } {
	    
	    set objType [lindex [lindex $object 0] 0] ; # Type of container
	    set objRef  [lindex [lindex $object 0] 1] ; # Object reference (Java handle)
	    set objSelected [lindex [lindex $object 0] 2] ; # Is container selected or not
	    
	} else {
	    
	    set objRef [lindex $object 1] ; # Object reference (Java handle)
	    set objSelected [lindex $object 2] ; # Is object part of selection or not
	    
	}	
	
	if {$objSelected} { set selString "SELECTED" } else { set selString "" }
	
	switch -regexp -- $objType {
	    
	    {^OB$} {

		#
		# Deal with single (loose) OBs
		#
		set execCommon(customPresetTime) [getTelescopePreset $objRef]
		set execCommon(rotatorOverhead) 0
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0
		setExecTimeForOneOB $objRef
		
	    }
	    
	    {GROUP|TIME_LINK} {
		
		#
		# Groups or Time-Links: deal with OBs individually
		#
		
		set execCommon(rotatorOverhead) 0
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0

		foreach obj [lindex $object 1] {
		    
		    set type [lindex $obj 0]
		    set ref [lindex $obj 1]
		    set selected [lindex $obj 2]

		    # We skip the OBs that were not explicitly selected
		    if {!$selected} {continue}

		    set execCommon(customPresetTime) [getTelescopePreset $ref]
		    setExecTimeForOneOB $ref

		}
				
	    }
	    
	    {CONCATENATION} {

		#
		# Concatenations: OBs executed back-to-back
		#

		set obCounter 0

		foreach obj [lindex $object 1] {

		    incr obCounter

		    set type [lindex $obj 0]
		    set ref [lindex $obj 1]
		    set selected [lindex $obj 2]

		    set execCommon(customPresetTime) [getTelescopePreset $ref]
    		    set execCommon(rotatorOverhead) 0
		    set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0

		    if { $obCounter > 1 } {

			# Compute preset time from OB <n> --> <n+1>
			set execCommon(customPresetTime) [getPresetTime $previousObRef $ref]

			# Compute rotator setup time from OB <n> --> <n+1>
			set execCommon(rotatorOverhead) [getRotatorOverhead $previousObRef $ref]

			# Compute filter change overhead between OB <n> and <n+1>
			set previousOBId [evmID $previousObRef]
			set OBId [evmID $ref]
			set execCommon(FILTER_CHANGE_BETWEEN_OBs) [computeFilterChangeBetweenOBs $previousOBId $OBId]
		    }


		    # Compute execution time only for selected OBs
		    if { $selected } {
			setExecTimeForOneOB $ref
		    }
		    
		    set previousObRef $ref
		}
		# Reset time to avoid counting it for the next concat.
		set execCommon(customPresetTime) 0
		set execCommon(rotatorOverhead) 0 
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0

	    }

	    default {
				
	    }
	}
	
    }

}

# Procedure: setSystemParameters
# arguments: none
# returns  : nothing
#
# Sets the system parameters
#
proc setSystemParameters {} {

    # reset the system parameters
    evmSystemResetParameters { "INS.FILT.NAME" "lastTemplateName" }
    
    # If running in BOB, then let the system parameters reflect
    # the state of the real system.
    if { [evmInBob] } {

	# set INS.FILT.NAME
	set value [bobDbRead "<alias>FILT.status(1)"]
	if { $value == "" } { set value "not-found" }
	evmSystemSetParameterValue "INS.FILT.NAME" $value
	
    }
}

# Call to the main procedure
setExecTime






